//---------------------------------------------------------------------------
//
//    bin2sym produces an encoded symbol number file representing a binary input file
//
//    Copyright (C) 2002  Barry D. Sanderson
//    Copyright (C) 2002  Carl A. Harcourt
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
//    Mailing addresses for the authors:
//
//    Barry D. Sanderson
//    1725 N. Bolton Ave.
//    Indianapolis, IN 46218  USA
//
//    Carl A. Harcourt
//    4937 W. 900 S.
//    Milroy, IN 46156  USA
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
//    Rev 30 Aug 2002 (CAH) - version 2.0, for initial PC/DOS release
//                            Developed using MS Visual C++ 6.0
//                            as a 'Win32 Console Application', ie a DOS .exe program
//                            Requires a Win32 environment for proper DOS execution
//                            This version will also build and execute unmodified on a Linux system
//
//    Rev 25 Nov 2001 (BDS) - version 1.1.0, for initial public release
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//  This program requires an input filename and optionally a numerical value
//  from the set {10, 20, 40, 70} as its first two arguments, in that order.
//  The output of this program is written to the file 'b2s_out.sym', and is a
//  sequence of symbol numbers, representing the encoded input file in the
//  message format defined below.
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//  Message format:
//
//      symbol        # of symbols      coments
//
//      lead_mod          74            used to allow receiver to measure transmitter's sample clock rate
//      id1xxx            1             1st of two symbols used to specify mode
//      id2xxx            1             2nd of two symbols used to specify mode
//        .
//        .               2             total of 3 [id1xxx id2xxx] pairs
//        .
//      frame1            1
//      frame2            1             4 symbol frame sequence delimits all code blocks
//      frame3            1
//      frame4            1
//      soh               1             start of header
//      nnnnn             1             length of file name, binary
//      fffff             2             length of file, binary LSB first
//      xxxxx             nnnnn         symbols for file name
//      sod               1             start of data
//      sssss             fffff         symbols for data in file
//        .
//        .                             after every data block the following 4 symbol sequence will be inserted:
//        .                                 frame1 frame2 frame3 frame4
//        .
//      ddddd             yy            dummy symbols to fill last block, as required
//      id1xxx            1             1st of two symbols used to specify mode
//      id2xxx            1             2nd of two symbols used to specify mode
//        .
//        .               2             total of 3 [id1xxx id2xxx] pairs
//        .
//      lead_mod          37            trailer to mark end of file
//
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// 'C' include files

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

//---------------------------------------------------------------------------
// application include files

#include "Wcommon.h"
#include "sys_def.h"
#include "sym_code.h"
#include "mode_def.h"
#include "rs_306.h"
#include "rd_sl_p.h"

//---------------------------------------------------------------------------
// variable declarations

char            cSignonMessage[]        = "Binary To Symbol Conversion Program";
char            cVersionMajor[]         = "R 02";
char            cVersionMinor[]         = "00";

unsigned int    rsmode;
unsigned int    bbs;
unsigned int    kxx [num_modes]         = {10, 20, 40, 70};

unsigned int    k1;
unsigned int    t1;
unsigned int    iidx_mn1;

SYM_TYPE_1      frame_buf [frame_len]   = {frame1, frame2, frame3, frame4};
SYM_TYPE_1      framer_buf [frame_len]  = {frame1r, frame2r, frame3r, frame4r};
SYM_TYPE_1      mode_buf [mode_len+mode_len];
SYM_TYPE_1      id1xx [num_modes]       = {id1x05, id1x10, id1x20, id1x40};
SYM_TYPE_1      id2xx [num_modes]       = {id2x05, id2x10, id2x20, id2x40};

SYM_TYPE_I      CodeBlock [SYMBOLS_PER_CODEBLOCK];                       // code block buffer

FILE          * FpInputFile;
FILE          * FpOutputFile;

char            cOutputFilename[]  = "b2s_out.sym";

int             iCodeBlockIndex;
int             fnlenmod;
int             iErrorNo;

SYM_TYPE_CH     frame_buf_ch [frame_len];
SYM_TYPE_CH     mode_buf_ch [mode_len+mode_len];
SYM_TYPE_CH     leader_buf_ch [leader_len];
SYM_TYPE_CH     sym_buf_ot [SYMBOLS_PER_CODEBLOCK];               // output symbol buffer

char            cMsg_noerr[]  = "Successful termination";
char            cMsg_ins[]    = "Usage:\n\n  bin2sym input_file num_of_redundancy_symbols (10|20|40|70) [ > output_file ]\n";
char            cMsg_ukner[]  = "Unknown error\n";
char            cMsg_noif[]   = "Couldn't open input file\n";
char            cMsg_noof[]   = "Call to stat() failed\n";

char            cMsg_05[]     = "Invalid symbol (out of range)\n";
char            cMsg_06[]     = "Unable to open default output file\n";
char            cMsg_07[]     = "";
char            cMsg_08[]     = "";
char            cMsg_09[]     = "Input file is too long\n";
char            cMsg_10[]     = "Memory allocation failed\n";
char            cMsg_11[]     = "Error reading input file\n";
char            cMsg_12[]     = "Error writing leader\n";
char            cMsg_13[]     = "Error writing body of file\n";
char            cMsg_14[]     = "";
char            cMsg_15[]     = "";
char            cMsg_16[]     = "";
char            cMsg_17[]     = "Unsupported number of redundancy symbols was requested\n";
char            cMsg_18[]     = "";
char            cMsg_19[]     = "Error writing frame sequence\n";

char *          erexmsg[]   = {cMsg_noerr, cMsg_ins,  cMsg_ukner, cMsg_noif, cMsg_noof, cMsg_05, cMsg_06, cMsg_07,
                               cMsg_08,  cMsg_09, cMsg_10,  cMsg_11, cMsg_12, cMsg_13, cMsg_14, cMsg_15, cMsg_16,
                               cMsg_17,  cMsg_18, cMsg_19};

//---------------------------------------------------------------------------
// function declarations

//---------------------------------------------------------------------------
int iWriteFrame (void)
{
    int   fwr;

    fwr = fwrite (frame_buf_ch, bytes_per_ch_sym, frame_len, FpOutputFile);
    if (fwr != frame_len)
        return -1;

    return 0;
}

//---------------------------------------------------------------------------
SYM_TYPE_CH   mk_ch_sym_num (SYM_TYPE_I ra[], SYM_TYPE_I ia[])
{

    //  calculate and return the integer corresponding to the concatenation of the base 9 values in ra[] and ia[]

    int             si;
    SYM_TYPE_CH     acum;

    acum = 0;
    for (si=NUM_INFO_SC-1; si>=0; si--)
        acum = (acum + ia[si]) * NUM_GF_EL_2;
    for (si=NUM_REDUNDANT_SC-1; si>0; si--)
        acum = (acum + ra[si]) * NUM_GF_EL_2;

    return (acum + ra[0]);
}

//---------------------------------------------------------------------------
int   iEncode (int *extra, int *info)
{
    //  extra   points to array where redundancy symbols of code word will be stored,
    //  info    points to array where information symbols have been stored

    //  returns:  0 if no errors were encountered
    //           -1 if an illegal symbol is found in the "info" array

    int     si, di;
    int     tmpi;

    for (si=0; si<(MAX_GF_EL_1 - 2*(int)t1); si++)
        if ((info[si] < 0) || (info[si] > MAX_GF_EL_1))
            return -1;
    for (di=0; di<(2*(int)t1); di++)
    {
        tmpi = 0;
        for (si=0; si<(MAX_GF_EL_1-2*(int)t1); si++)
            tmpi += (info[si]*rgm1xx[rsmode][di][si])%NUM_GF_EL_1;
        extra[di]= tmpi%NUM_GF_EL_1;
    }

    return 0;
}

//---------------------------------------------------------------------------
int iProcessBlock (void)
{
    //  iEncode contents of CodeBlock[],
    //  encode2 each symbol in CodeBlock[]
    //  write channel symbol numbers to sym_buf_ot,
    //  fwrite result to output file
    //
    //  return   0 on success,
    //          -1 on failure

    int         j, p;
    int         i, er;


    i = iEncode(&CodeBlock[0], &CodeBlock[iidx_mn1]);
    if (i!=0)
        {iErrorNo = 5; return -1;}
    for (j=0; j<SYMBOLS_PER_CODEBLOCK; j++)
    {
        for (p=0; p<NUM_INFO_SC; p++)
            ch_sym_buf_i[p] = sym_to_phi[CodeBlock[MAX_INFO_INDEX-j]][p];
        er = encode2(&ch_sym_buf_r[0], &ch_sym_buf_i[0]);
        if (er !=0)
        {
            iErrorNo = 13;
            return -1;
        }
        sym_buf_ot[j] = mk_ch_sym_num(&ch_sym_buf_r[0], &ch_sym_buf_i[0]);
    } // end of for j

    i = fwrite(sym_buf_ot, bytes_per_ch_sym, SYMBOLS_PER_CODEBLOCK, FpOutputFile);
    if (i!=SYMBOLS_PER_CODEBLOCK)
    {
        iErrorNo = 13;
        return -1;
    }

    return 0;
}

//---------------------------------------------------------------------------

int main (int argc, char *argv[], char *env[])
{

    //-----------------------------------------------------------------------
    // main() variable declarations

    unsigned long      si, di;
    unsigned long      ulByteCount;
    unsigned long      ulFileLength, ulBytesUsed, ulBytesLeft;
    unsigned long      ulFilenameLength, ir, fwr, tl, sl, bl;

    struct stat        statb;
    int                statr;

    unsigned char      ucFilenameBuffer [MAX_FILENAME_LENGTH+2];
    unsigned char    * ucpFileBuffer;

    SYM_TYPE_CH        csn;

    //-----------------------------------------------------------------------
    // main() begins here

    //  arv[1] is input file name
    //  arv[2] is mode value

    //-----------------------------------------------------------------------
    // print sign on message
    fprintf (stdout, "\n%s - %s.%s\n\n", cSignonMessage, cVersionMajor, cVersionMinor);

    //-----------------------------------------------------------------------
    // check command line

    if (argc < 3)       // enough arguments?
    {
        fprintf (stderr, erexmsg[1]);
        exit (1);
    }

    //-----------------------------------------------------------------------
    // get input file info

    statr = stat(argv[1], &statb);
    if (statr!=0)
       {iErrorNo=4 ; goto exit_a;}

    ulFileLength = (unsigned long) (statb.st_size);
    if (ulFileLength > MAX_FILE_LENGTH)
       {iErrorNo=9; goto exit_a;}

    ulFilenameLength = strlen(argv[1]);
    if (ulFilenameLength > MAX_FILENAME_LENGTH)
    {
        fprintf(stderr,"WARNING: filename is being limited to %3u characters\n", MAX_FILENAME_LENGTH);
        ulFilenameLength=MAX_FILENAME_LENGTH;
    }

    //-----------------------------------------------------------------------

    si = atoi(argv[2]);             // get mode value
    di = 0;
    while (di < num_modes)
    {
        if (si != kxx[di])
            di++;
        else                        // set mode related values
        {
            rsmode   = di;
            k1       = kx[di];
            t1       = k1/2;
            iidx_mn1 = iidx_mnx[rsmode];

            for (ir = 0; ir < 2*mode_len; ir += 2)
                mode_buf[ir]=id1xx[di];
            for (ir = 1; ir < 2*mode_len; ir += 2)
                mode_buf [ir] = id2xx[di];

            bbs      = MAX_GF_EL_1 - k1;
            goto modeok;
        }
    }
    fprintf (stderr, erexmsg[17]);
    exit (17);

    //-----------------------------------------------------------------------

modeok:

    //-----------------------------------------------------------------------
    // open default output file

    remove (cOutputFilename);                         // delete existing file
    FpOutputFile = NULL;
    FpOutputFile = fopen (cOutputFilename, "wb");      // open default output file
    if (FpOutputFile == NULL)
    {
        iErrorNo = 6;
        goto exit_b;
    }

    //-----------------------------------------------------------------------
    // initialize leader_buf_ch
    for (si=0; si< NUM_INFO_SC; si++)
        ch_sym_buf_i[si]= sym_to_phi[lead_i][si];
    for (si=0; si< NUM_REDUNDANT_SC; si++)
        ch_sym_buf_r[si]= sym_to_phi[lead_r][si];

    csn = mk_ch_sym_num(&ch_sym_buf_r[0], &ch_sym_buf_i[0]);
    for (si=0; si<leader_len; si++)
        leader_buf_ch[si]=csn;

    //-----------------------------------------------------------------------
    // initialize frame_buf_ch
    for (si=0; si< NUM_INFO_SC; si++)
        ch_sym_buf_i[si]= sym_to_phi[frame1][si];
    for (si=0; si< NUM_REDUNDANT_SC; si++)
        ch_sym_buf_r[si]= sym_to_phi[frame1r][si];
    frame_buf_ch[0]= mk_ch_sym_num(&ch_sym_buf_r[0], &ch_sym_buf_i[0]);

    for (si=0; si< NUM_INFO_SC; si++)
        ch_sym_buf_i[si]= sym_to_phi[frame2][si];
    for (si=0; si< NUM_REDUNDANT_SC; si++)
        ch_sym_buf_r[si]= sym_to_phi[frame2r][si];
    frame_buf_ch[1]= mk_ch_sym_num(&ch_sym_buf_r[0], &ch_sym_buf_i[0]);

    for (si=0; si< NUM_INFO_SC; si++)
        ch_sym_buf_i[si]= sym_to_phi[frame3][si];
    for (si=0; si< NUM_REDUNDANT_SC; si++)
        ch_sym_buf_r[si]= sym_to_phi[frame3r][si];
    frame_buf_ch[2]= mk_ch_sym_num(&ch_sym_buf_r[0], &ch_sym_buf_i[0]);

    for (si=0; si< NUM_INFO_SC; si++)
        ch_sym_buf_i[si]= sym_to_phi[frame4][si];
    for (si=0; si< NUM_REDUNDANT_SC; si++)
        ch_sym_buf_r[si]= sym_to_phi[frame4r][si];
    frame_buf_ch[3]= mk_ch_sym_num(&ch_sym_buf_r[0], &ch_sym_buf_i[0]);

    //-----------------------------------------------------------------------
    // initialize mode_buf_ch
    for (si=0; si< NUM_INFO_SC; si++)
        ch_sym_buf_i[si]= sym_to_phi[mode_buf[0]][si];
    ir = encode2(&ch_sym_buf_r[0], &ch_sym_buf_i[0]);
    if (ir != 0)
        {iErrorNo=5 ; goto exit_a;}

    csn = mk_ch_sym_num(&ch_sym_buf_r[0], &ch_sym_buf_i[0]);
    for (si=0; si<mode_len+mode_len; si+=2)      /*??*/
        mode_buf_ch[si]=csn;
    for (si=0; si< NUM_INFO_SC; si++)
        ch_sym_buf_i[si]= sym_to_phi[mode_buf[1]][si];
    ir = encode2(&ch_sym_buf_r[0], &ch_sym_buf_i[0]);
    if(ir!=0)
        {iErrorNo=5 ; goto exit_a;}

    csn = mk_ch_sym_num(&ch_sym_buf_r[0], &ch_sym_buf_i[0]);
    for (si=1; si<mode_len+mode_len; si+=2) /*??*/
        mode_buf_ch[si]=csn;

    //-----------------------------------------------------------------------
    strncpy(ucFilenameBuffer, argv[1], ulFilenameLength);
    ucFilenameBuffer[ulFilenameLength]   = 0x00;
    ucFilenameBuffer[ulFilenameLength+1] = 0x00;
    ucpFileBuffer = (unsigned char *) malloc((unsigned long) (ulFileLength+2)*sizeof(char));
    if (ucpFileBuffer==NULL)
        {iErrorNo=10; goto exit_a;}

    //----------------------------------------------------------------------
    FpInputFile=fopen(argv[1], "rb");
    if (FpInputFile==NULL)
        {iErrorNo=3; goto exit_b;}

    ir = fread(ucpFileBuffer, 1, ulFileLength, FpInputFile);
    if (ir!=ulFileLength)
        {iErrorNo=11 ; goto exit_c;}
    if (ferror(FpInputFile) != 0)
        {iErrorNo=11 ; goto exit_c;}

    ucpFileBuffer[ulFileLength]   = 0x00;
    ucpFileBuffer[ulFileLength+1] = 0x00;

    //-----------------------------------------------------------------------
    // write leader into output file

    fwr = fwrite (leader_buf_ch, bytes_per_ch_sym, leader_len, FpOutputFile);
    if (fwr!=leader_len)
        {iErrorNo=12; goto exit_c;}

    //-----------------------------------------------------------------------
    // put mode specifying sequence into output file
    ir  = mode_len+mode_len;
    fwr = fwrite(mode_buf_ch, bytes_per_ch_sym, ir, FpOutputFile);
    if (fwr!=ir)
        {iErrorNo=12; goto exit_c;}

    //-----------------------------------------------------------------------
    // put fixed-length part of header into CodeBlock[]
    fwr = iWriteFrame();
    if (fwr !=0)
        {iErrorNo=19; goto exit_c;}

    iCodeBlockIndex       = MAX_INFO_INDEX;
    CodeBlock[iCodeBlockIndex--] = (unsigned int) (soh);
    CodeBlock[iCodeBlockIndex--] = (unsigned int) (ulFilenameLength);
    CodeBlock[iCodeBlockIndex--] = (unsigned int) (ulFileLength & 0xff);
    CodeBlock[iCodeBlockIndex--] = (unsigned int) ((ulFileLength>>8) & 0xff);

    //-----------------------------------------------------------------------
    // put file name into header
    for (si=0; si<ulFilenameLength; si++)
        CodeBlock[iCodeBlockIndex--]=(unsigned int)(ucFilenameBuffer[si]);

    //-----------------------------------------------------------------------
    // put start of data into code block
    CodeBlock[iCodeBlockIndex--] = (unsigned int) sod;
    ulBytesLeft = ulFileLength;
    ulBytesUsed = 0;
    tl          = 0;
    ulByteCount = 0;

    //-----------------------------------------------------------------------
    // do bulk of data from file
    while (ulBytesLeft > bbs)
    {
        sl = iCodeBlockIndex-iidx_mn1+1;
        bl = sl;
        for (si=0; si<bl; si++)
            CodeBlock[iCodeBlockIndex--] = (unsigned int) ucpFileBuffer[ulBytesUsed+si];
            if (iProcessBlock()!=0)
                goto exit_c;
        ulBytesUsed     += (unsigned long) bl;
        ulBytesLeft     -= (unsigned long) bl;
        iCodeBlockIndex  = MAX_INFO_INDEX;
        ulByteCount++;
        if (ulByteCount == blks_per_frame)
        {
            ulByteCount = 0;
            fwr  = iWriteFrame();
            if (fwr!=0)
                {iErrorNo=19; goto exit_c;}
        }
    } // end of while ulBytesLeft > bbs

    //-----------------------------------------------------------------------
    // do rest of data from input file
    while (ulBytesLeft > 0)
    {
        sl = iCodeBlockIndex - iidx_mn1 + 1;
        bl = sl;
        if (ulBytesLeft >= bl)
        {
            for (si=0; si<bl; si++)
                CodeBlock[iCodeBlockIndex--] = (unsigned int)(ucpFileBuffer[ulBytesUsed+si]);
            if (iProcessBlock()!=0)
                goto exit_c;
            ulBytesUsed += bl;
            ulBytesLeft -= bl;
        } // end of ulBytesLeft >= bl
        else
        {
            if (ulBytesLeft > 0)
            {
                for (si=0; si<ulBytesLeft; si++)
                    CodeBlock[iCodeBlockIndex--] = (unsigned int) ucpFileBuffer[ulBytesUsed+si];
                ulBytesUsed += ulBytesLeft;
                ulBytesLeft  = 0;
                sl           = iCodeBlockIndex - iidx_mn1+1;
                for (si=0; si<sl; si++)
                    CodeBlock[iCodeBlockIndex--]= (unsigned int)(dummy1);
                tl += sl;
                if (iProcessBlock()!=0)
                    goto exit_c;
            } // end of ulBytesLeft > 0
        } // end of ulBytesLeft < bl

        iCodeBlockIndex = MAX_INFO_INDEX;
        ulByteCount++;
        if(ulByteCount==blks_per_frame)
        {
            ulByteCount = 0;
            fwr  = iWriteFrame();
            if (fwr != 0)
                {iErrorNo=19; goto exit_c;}
        }
    } // end of while ulBytesLeft > 0

    //-----------------------------------------------------------------------
    // put mode specifying sequence into output file

    ir  = mode_len + mode_len;
    fwr = fwrite (mode_buf_ch, bytes_per_ch_sym, ir, FpOutputFile);
    if (fwr != ir)
        {iErrorNo=12; goto exit_c;}

    //-----------------------------------------------------------------------
    // put trailer symbols into output file
    fwr = fwrite (leader_buf_ch, bytes_per_ch_sym, trailer_len, FpOutputFile);
    if (fwr != trailer_len)
        {iErrorNo=12; goto exit_c;}

    //-----------------------------------------------------------------------
    // exeunt all

    iErrorNo = 0;        // got this far, no errors

exit_c:
    fclose (FpInputFile);
    fclose (FpOutputFile);

exit_b:
    free (ucpFileBuffer);

exit_a:
    if (iErrorNo != 0)
        fprintf(stderr, erexmsg[iErrorNo]);

    exit(iErrorNo);
}
