//---------------------------------------------------------------------------
//
//    DFT routines used by Wyman digital communications programs
//
//    Fast Fourier Transform on an integer-power-of-2 number of
//    complex floating point data values.
//
//    Copyright (C) 2002  Barry D. Sanderson
//    Copyright (C) 2002  Carl A. Harcourt
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
//    Mailing addresses for the authors:
//
//    Barry D. Sanderson
//    1725 N. Bolton Ave.
//    Indianapolis, IN 46218  USA
//
//    Carl A. Harcourt
//    4937 W. 900 S.
//    Milroy, IN 46156  USA
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
//    Rev 30 Aug 2002 (CAH) - version 2.0, for initial PC/DOS release
//                            Developed using MS Visual C++ 6.0
//                            as a 'Win32 Console Application', ie a DOS .exe program
//                            Requires a Win32 environment for proper DOS execution
//                            This version will also build and execute unmodified on a Linux system
//
//    Rev 25 Nov 2001 (BDS) - version 1.1.0, for initial public release
//    Rev 17 Dec 1992 (BDS) - Function to peform foward or inverse Discrete
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// 'C' include files

#include <stddef.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <float.h>

//---------------------------------------------------------------------------
// application include files

#include "Wcommon.h"
#include "fft.h"
#include "dcom.h"

//---------------------------------------------------------------------------
// global variable declarations

//---------------------------------------------------------------------------
// local variable declarations

//---------------------------------------------------------------------------
// static function prototypes

//---------------------------------------------------------------------------
// public function definitions


//---------------------------------------------------------------------------
double  init_dft3 (unsigned int dft_size, float t_step, float * quarter_sine_tbl)
{
    unsigned int    i;
    double          s_step;

    // calculate quarter cycle sine table and return frequency step

    s_step = (double) PI2 / (double) dft_size;

    for (i=0; i<=(dft_size/4); i++)
    {
        quarter_sine_tbl[i] = (float) sin (s_step * (double) i);
    }

    return (double) (1.0/(t_step * (float) dft_size));
}

//---------------------------------------------------------------------------
void    dfft3f (float *      fInputReal,
                float *      fInputImaginary,
                float *      fOutputReal,
                float *      fOutputImaginary,
                float *      fSineTable,
                unsigned int uiIndexMax,
                unsigned int uiInverseDFTFlag,
                unsigned int uiRealInputFlag)
{
    //  fInputReal          points to real part of input
    //  fInputImaginary     points to imaginary part of input
    //  fOutputReal         points to real part of result
    //  fOutputImaginary    points to imaginary part of result
    //
    //  fSineTable          points to a quarter period sine table  [0:(uiIndexMax+1)/4]
    //  uiIndexMax          maximum index of arrays: fInputReal, fInputImaginary, fOutputReal, and fOutputImaginary [0:uiIndexMax]
    //  uiInverseDFTFlag    1 if inverse DFT is to be computed
    //  uiRealInputFlag     1 if input imaginary part is 0

    float           fRr, fRi, fR1r, fR1i, fS1, fS2;
    unsigned long   i, j, k, l, m;
    unsigned long   NT, NH, NQ, IR, IT;
    unsigned long   ID, ND;

    // copy input data array to output array
    // setting imaginary part to zero, if input is not complex

    for (i=0; i<=uiIndexMax; i++)
        fOutputReal[i] = fInputReal[i];

    if ((uiRealInputFlag == 1) && (uiInverseDFTFlag != 1))
    {
        for (i=0; i<=uiIndexMax; i++)
             fOutputImaginary[i] = 0.0;
    }
    else
    {
        for (i=0; i<=uiIndexMax; i++)
             fOutputImaginary[i] = fInputImaginary[i];
    }

    NT = uiIndexMax;
    NQ = NT/4 + 1;
    NH = NQ+NQ;

    // reorder initial data, so result will be in desired order
    j = 0;
    for (i=0; i<=(NT-2); i++)
    {
        if (j>i)
        {
            fRr                 = fOutputReal[j];
            fRi                 = fOutputImaginary[j];
            fOutputReal[j]      = fOutputReal[i];
            fOutputImaginary[j] = fOutputImaginary[i];
            fOutputReal[i]      = fRr;
            fOutputImaginary[i] = fRi;
        }
        k                   = NH;
        fOutputImaginary[i] = fInputImaginary[i];
        while (j>=k)
        {
            j = j-k;
            k = k/2;
        }
        j = j+k;
    }

    // calculate the discrete Fourier transform
    ID = 1;
    IR = 1;
    m  = 0;
    IT = 0;

tran:

    do
    {
        i  = ID;
        ID = ID + ID;
        for (j=0; j<i; j++)
        {
            fS2 = fSineTable[m];
            if (uiInverseDFTFlag != 1)
                fS2 = -fS2;
            fS1 = fSineTable[NQ-m];
            if (j>=IR)
            {
                m   = m - IT;
                fS1 = -fS1;
            }
            else
                m = m + IT;
            for (ND=j; ND<=NT; ND+=ID)
            {
                k                   = ND;                 // ul assigned to ui?
                l                   = k + i;
                                                            // butterfly computation
                fR1r                = fOutputReal[l];
                fR1i                = fOutputImaginary[l];
                fRr                 = (fS1 * fR1r) - (fS2 * fR1i);
                fRi                 = (fS1 * fR1i) + (fS2 * fR1r);
                fR1r                = fOutputReal[k];
                fR1i                = fOutputImaginary[k];
                fOutputReal[k]      = fR1r + fRr;
                fOutputImaginary[k] = fR1i + fRi;
                fOutputReal[l]      = fR1r - fRr;
                fOutputImaginary[l] = fR1i - fRi;
            }
        }

        IR = i;
        IT = NQ/i;
    }
    while (i < NH);

    if (uiInverseDFTFlag != 1)
        return;

    fRr = (float) (uiIndexMax + 1);
    for (i=0; i<=uiIndexMax; i++)
    {
        fOutputReal[i]      /= fRr;
        fOutputReal[i]      /= fRr;
        fOutputImaginary[i] /= fRr;
        fOutputImaginary[i] /= fRr;
    }
    return;
}
