//---------------------------------------------------------------------------
//
//    wav to binary conversion program
//
//    demodulator-decoder program for Wyman digital communications method
//
//    Copyright (C) 2002  Barry D. Sanderson
//    Copyright (C) 2002  Carl A. Harcourt
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
//    Mailing addresses for the authors:
//
//    Barry D. Sanderson
//    1725 N. Bolton Ave.
//    Indianapolis, IN 46218  USA
//
//    Carl A. Harcourt
//    4937 W. 900 S.
//    Milroy, IN 46156  USA
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
//    Rev 03 Oct 2002 (BDS) - limited dmd_fidx to non-negative values
//
//    Rev 24 Sep 2002 (BDS) - version 2.01, added: detection of wrong number of bytes
//                            written to disk files; freeing of dmamp[][];
//                            closing of irf2 for certain error exits;
//                            removing scratch files.
//                            changed value  of gap_offi, due to some
//                            subcarriers being modulated during leader and
//                            trailer; relaxed criteria for "good" subcarrier
//                            and for success in lm_match3; relaxed criteria
//                            for haveing found leader and trailer.
//                            fixed array of "good" subcarrier values passed
//                            to "iCalcBestFit" (previously this was wrong unless
//                            subcarriers were "good"
//
//    Rev 30 Aug 2002 (CAH) - version 2.00, for initial PC/DOS release
//                            Developed using MS Visual C++ 6.0
//                            as a 'Win32 Console Application', ie a DOS .exe program
//                            Requires a Win32 environment for proper DOS execution
//                            This version will also build and execute unmodified on a Linux system
//
//
//    Rev 10 Dec 2001 (BDS) - changed time offset between demodulated, filtered,
//                            angle differences, and demodulated, average amplitudes; dbgl expanded
//                            dblg = 35 introduced, for control of information dumped to stderr;
//                            changed    dumping of demodulated signals so that 40< dbgl <50 dumps
//                            demodulated angle differences, and dbgl >= 50 dumps demodulated
//                            amplitudes
//    Rev 09 Dec 2001 (BDS) - added dump of results from adjusting symbol positions
//                            based on amplitude information to stderr
//    Rev 05 Dec 2001 (BDS) - percentage of reserved erasures (era_rsv[]) doubled
//    Rev 25 Nov 2001 (BDS) - cosmetic changes, functionallity not changed
//    Rev 21 Nov 2001 (BDS) - inner decoding processed augmented to take
//                            care of some cases that actually have more than 2 errors, but
//                            are "close enough" with only 1 or 2 subcarriers erased
//    Rev 20 Nov 2001 (BDS) - fixes to addition of 15 Nov 2001
//    Rev 15 Nov 2001 (BDS) - addition of compensation for rapid variations
//                            in propogation delay, using variations in amplitude,
//                            format of dbgl > 5 changed
//    Rev 18 Aug 2001 (BDS) - debuging of new inner decoding proceedure
//    Rev 05 Aug 2001 (BDS) - new decoding proceedure for inner code
//    Rev 16 Jun 2001 (BDS) - low pass filter used instead of averaging, on
//                            normalized angle differences
//    Rev 12 Jun 2001 (BDS) - use of actual framing sequence positions, and best
//                            fit of these positions, for best subcarrier added between 1st
//                            and 2nd pass through demodulator
//    Rev 05 Jun 2001 (BDS) - fine tuning bypassed, 2 demodulation passes forced
//    Rev 02 Jun 2001 (BDS) - introduction of :
//                            interpolation of narrowband filtered data (to get fine tuning slope
//                            and interecpt)
//                            2nd pass through demodulator
//                            Replacement of numerically unstable correlation coefficient calculation
//                            in determining which fine tuing data to use
//                            use of only "good enough" fine tuning data
//    Rev 01 Jun 2001 (BDS) - changed version 1.0.0 of pm7b for Wyman1x (pm8a) case
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// 'C' include files

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

//---------------------------------------------------------------------------
//  application include files

#include "Wcommon.h"
#include "dcom_t.h"
#include "decode.h"
#include "fft.h"
#include "filter.h"
#include "frame.h"
#include "gf_307.h"
#include "mode_def.h"
#include "ref_spec.h"
#include "sym_code.h"

//---------------------------------------------------------------------------
// defines

#define W2B_INPUT_VAR_TYPE  short int

#define DFT_SIZE        4096
#define DFT_SIZE_W      (DFT_SIZE/4)
#define dmoff           (DFT_SIZE/2)
#define dmoffW          (DFT_SIZE_W/2)
#define wavhdrlenb      44L                 // .wav header length to skip
#define clk_freq_nom    11025.0
#define dumd            1                   // dummy data value
#define delay_nom       90                  // clock periods per symbol
#define ave_nom         21                  // number of values to average together
#define NASN            0x200               // for 1 byte per symbol case
#define cppsym          delay_nom           // clock periods per symbol, nominal
#define cppblk_nom      ((MAX_GF_EL_1+frame_len)*cppsym) // clock periods per block, nominal
#define num_modes_l     1                   // number of modes (different leaders) supported by this program
#define num_files       10

// for + and - 40 Hz around nominal frequencies, 1024 point DFT
#define tsi_start       42                  // 500 - 40 Hz
#define tsi_count       9                   // 97 Hz range

// for + and - 200 Hz around nominal frequencies
#define lsi_startW      27                  // 500 - 200 Hz
#define lsi_countW      37                  // 400 Hz range
#define lsi_centerW     46                  // index of start of pattern, for no shift
#define l1_FOUND        0x01
#define t1_FOUND        0x02
#define frsq_srsym      60                  // # of symbol periods over which to search for framing sequence
#define frsq_srsam      (frsq_srsym * cppsym)

    // # of sample periods over which to search for framing sequence
#define frosrng         42                  // one sided acceptance range for initial framing sequence

#define frosftr         34                  // one sided search range for fine tuning of framing sequence ranges
#define frosftr2        72                  // one sided search range for fine tuning of framing sequence ranges, version 2
#define cppblk_ossr     26                  // one sided search range for fine tuning of cppblk
#define gap_offi        8                  // number of 4096 DFT points from center
                                            // to measurement point between sub-carriers
#define fdthld_nom      0.01                // threshold for sub-carrier spacing ratio
#define xcthld          0.6                 // minimum cross-correlation value to accept
#define nsc_min         5                   // minimum number of subcarriers required to be present
#define ossr2           6                   // one sided search range around calculated positions of local maximums, from cross-correlation
#define mode1dthld      13.000              // distance threshold for mode 1, no erasures
#define mode2dthld      13.60147            // distance threshold for mode 2, no erasures
#define slope_max       1.01                // 10,000 parts per million
#define slope_min       0.99
#define off_hz_max      400.
#define off_hz_min      -400.
#define cppblk_max      28179               // 10,000 parts per million
#define cppblk_min      27621
#define cppblk_off_lim  278
#define max_usr_sym     0xff
#define mseavexf        1.0                 // factor to multiply average mean squared err
                                            // by, to get threshold for good/bad decision on individual framing sequences
#define ossi_amp1       18                  // one sided search range for initial symbol position refinement search
#define thld_amp1       0.6                 // threshold value for cross-correlation value in initial symbol position refinement
#define ossi_amp2       18                  // one sided search range for 2nd pass symbol position refinement search
#define thld_amp2       0.5                 // threshold value for cross-correlation value in 2nd pass symbol position refinement

#define VOIDARRAY           0               // for array display function
#define CHARARRAY           1
#define UNSIGNEDCHARARRAY   2
#define INTARRAY            3
#define UNSIGNEDINTARRAY    4
#define LONGARRAY           5
#define UNSIGNEDLONGARRAY   6
#define FLOATARRAY          7
#define DOUBLEARRAY         8
#define DEFAULTARRAY        0

#define DEFAULTDEBUGLEVEL   1

//---------------------------------------------------------------------------
// external variable declarations

extern  int             errno;

//---------------------------------------------------------------------------
// variable declarations

//---------------------------------------------------------------------------
unsigned int    uiDebugLevel = DEFAULTDEBUGLEVEL;

//  uiDebugLevel value controls what information is printed to stdout, stderr, and w2b_dbg.txt
//
//  (uiDebugLevel >   0) => final tuning, final cppblk, mode weights, and error counts written
//  (uiDebugLevel >  10) => leader & trailer information and initial framing sequence search
//                         information written
//  (uiDebugLevel >  20) => final framing sequence search information written
//  (uiDebugLevel == 25) => DOS version debug
//  (uiDebugLevel >  30) => positions for all framing sequences written
//  (uiDebugLevel >  35) => symbol position adjustment information, based on amplitude,
//                         written to stderr
//  (uiDebugLevel >  40) => gnuplot style files containing demodulation results
//                         for each subcarrier written; 40 < uiDebugLevel < 50 dumps
//                         filtered, normalized angle differences; uiDebugLevel >= 50
//                         dumps average amplitudes
//  (uiDebugLevel >  50) => symbol numbers for both codes for each block written
//---------------------------------------------------------------------------

float                   t_step;                 // nominally 90.702947e-6 for 11025 Hz

int                     soff [NUM_TOTAL_SC] =
                                {soff0, soff1, soff2, soff3, soff4, soff5, soff6, soff7};

unsigned int            eraii [NUM_TOTAL_SC];

unsigned int            k1;
unsigned int            t1;
unsigned int            iidx_mn1;

#if (s0blen > s0nb2blen)
W2B_INPUT_VAR_TYPE      samsi [s0blen];
#else
W2B_INPUT_VAR_TYPE      samsi [s0nb2blen];
#endif

W2B_INPUT_VAR_TYPE      datarI [DFT_SIZE];

float                   fInputReal      [DFT_SIZE];
float                   fInputImag      [DFT_SIZE];
float                   fOutputReal     [DFT_SIZE];
float                   fOutputImag     [DFT_SIZE];
float                   mag     [3]     [DFT_SIZE];
float                   wind            [DFT_SIZE];
float                   windW           [DFT_SIZE_W];
float                   snqt            [(DFT_SIZE/4)+1];
float                   snqtW           [(DFT_SIZE_W/4)+1];

unsigned long           ulSourceIndex;
unsigned long           ulHeaderLengthB;
unsigned long           sstepW, xctstart;
unsigned int            rsmode, bbs;
int                     val_read, val_to_write;

FILE    *               FpInputFile;
FILE    *               FpOutputFile;
FILE    *               FpScratchFile1;
FILE    *               FpScratchFile2;
FILE    *               FpDebugFile;         // 2002 apr19

char    *               cpOutputFilename;
char                    cDefaultOutputFilename[] = "wav2bin.out";
char                    cDebugFilename[]         = "w2b_dbg.txt";

SYM_TYPE_CH             ch_sym_buf[SYMBOLS_PER_CODEBLOCK];       // output symbol buffer
unsigned int            cb[2][SYMBOLS_PER_CODEBLOCK];            // code block buffers

float                   ct_slope, ct_intercept, f_slope, f_intercept;

// following is array of phase modulation center frequencies
double          pm_nom_ctr[NUM_TOTAL_SC] =
                        {pmc0, pmc1, pmc2, pmc3, pmc4, pmc5, pmc6, pmc7};

#define    num_sc_l    12

// following is array of nominal frequencies for leader and trailer
double          pm_nom_lt [num_sc_l] =
                {
                    pmc0, pmc1, 1006.167, 1128.684, 1236.168, 1358.666,
                    1401.326, 1523.835, 1631.327, 1753.832, pmc6, pmc7
                };

#if (fap > fapnb2)
float           Ssbc[fap], Ssbs[fap], Amp[fap], buf2[fap];
#else
float           Ssbc[fapnb2], Ssbs[fapnb2], Amp[fapnb2], buf2[fapnb2];
#endif

#if (s0blen > s0nb2blen)
float           Sambuf[s0blen], Lo1c[s0blen], Lo1s[s0blen];
#else
float           Sambuf[s0nb2blen], Lo1c[s0nb2blen], Lo1s[s0nb2blen];
#endif

#if (s1blen > s1nb2blen)
float           Lo1cf[s1blen], Lo1sf[s1blen];
float           Lo2cc[s1blen], Lo2cs[s1blen], Lo2ss[s1blen], Lo2sc[s1blen];
#else
float           Lo1cf[s1nb2blen], Lo1sf[s1nb2blen];
float           Lo2cc[s1nb2blen], Lo2cs[s1nb2blen], Lo2ss[s1nb2blen], Lo2sc[s1nb2blen];
#endif

#if (s2blen > s2nb2blen)
float           Lo2cs1[s2blen], Lo2ss1[s2blen];
#else
float           Lo2cs1[s2nb2blen], Lo2ss1[s2nb2blen];
#endif

#if (s3blen > s3nb2blen)
float           Lo2cs2[s3blen], Lo2ss2[s3blen], Ampb[s3blen];
#else
float           Lo2cs2[s3nb2blen], Lo2ss2[s3nb2blen], Ampb[s3nb2blen];
#endif

// following are file names used for demodulated output values for each subcarrier

unsigned char   sb0db[] = "lo0dif.gnu";
unsigned char   sb1db[] = "lo1dif.gnu";
unsigned char   sb2db[] = "lo2dif.gnu";
unsigned char   sb3db[] = "lo3dif.gnu";
unsigned char   sb4db[] = "lo4dif.gnu";
unsigned char   sb5db[] = "lo5dif.gnu";
unsigned char   sb6db[] = "lo6dif.gnu";
unsigned char   sb7db[] = "lo7dif.gnu";
unsigned char   sb8db[] = "lo8dif.gnu";
unsigned char   sb9db[] = "lo9dif.gnu";

unsigned char * lodif7nfn[num_files] =
                {
                    sb0db, sb1db, sb2db, sb3db, sb4db, sb5db, sb6db, sb7db, sb8db, sb9db
                };

float         * dif7n [NUM_TOTAL_SC];
float         * dmamp [NUM_TOTAL_SC];
float           ec_mode_dist [num_ec_lvls];

//      NOTE the two arrays below MUST be set to agree with the
//      symbol values actually used for these modes

SYM_TYPE_I      ec_mode_id1xxsym [num_ec_lvls] = {12,  6, 10,  3};
SYM_TYPE_I      ec_mode_id2xxsym [num_ec_lvls] = { 6,  5,  9, 10};

int             ec_mode;
float           ref_a;

int             exidx [num_sc_l];
int             exidxW [num_sc_l];

int             frsqbioff [NUM_TOTAL_SC];
float           frsqbmse [NUM_TOTAL_SC];
double          f_stepd, f_stepdW;

struct          lxcd2
        {
            unsigned long   sidx;
            float           r53;
            float           off_hz;
            float           slope;
            unsigned int    state;
            unsigned int    mode;
            int             offi;
            unsigned int    nlm;
            float           xcv;
            unsigned int    lmbm;
            float           lmfreq[num_sc_l];
            float           lmqual[num_sc_l];
            unsigned int    lmidx[num_sc_l];
            unsigned int    lmgoodi[num_sc_l];
            int             lmcount;
        };

int             midx [3];
struct lxcd2    lead_x[3];

char    cSignonMessage[]    = "Wav To Binary Conversion Program";
char    cVersionMajor[]     = "R 02";
char    cVersionMinor[]     = "01";

char    cUsage[]   = "Usage:\n\n  wav2bin datafile.{wav|sw} [DebugLevel(nn)] [DefaultMode(0-3)] [SampleFreq(Hz)]";

char    cError00[] = "Successful termination";
char    cError02[] = "malloc could not allocate enough memory.";
char    cError03[] = "Unknown error";
char    cError04[] = "fseek failed";
char    cError05[] = "WRONG number of bytes written to disk file";
char    cError06[] = "stat failed";
char    cError07[] = "Couldn't open file for input";
char    cError08[] = "Couldn't open file for output";
char    cError09[] = "";
char    cError10[] = "The leader was not found in the first half of the input file";
char    cError11[] = "The trailer was not found in the last half of the input file";
char    cError12[] = "leader too corrupted for fine tuning operations";
char    cError13[] = "error trying to read from scratch file";
char    cError14[] = "logic error in step 3 of symbol position refinement";
char    cError15[] = "logic error in step 4 of symbol position refinement";
char    cError16[] = "";
char    cError17[] = "";
char    cError18[] = "";
char    cError19[] = "Start of Header symbol not in proper place";
char    cError20[] = "recovered file name length is too long";
char    cError21[] = "recovered file name length is too short";
char    cError22[] = "recovered file length is too long";
char    cError23[] = "illegal symbol in file name";
char    cError24[] = "Start of Data symbol not in proper place";
char    cError25[] = "illegal symbol in body of file";
char    cError26[] = "error while writing output file";
char    cError27[] = "last codeblock(s) ignored: likely due to incorrectly recovered file length";
char    cError28[] = "";
char    cError29[] = "";
char    cError30[] = "";
char    cError31[] = "Sample Frequency is out of range\n";
char    cError32[] = "Digital filter could not allocate memory\n";
char    cError33[] = "";

char *  erexmsg[] = {
                     cError00, cUsage,   cError02, cError03, cError04, cError05, cError06,
                     cError07, cError08, cError09, cError10, cError11, cError12, cError13, cError14, cError15,
                     cError16, cError17, cError18, cError19, cError20, cError21, cError22, cError23, cError24,
                     cError25, cError26, cError27, cError28, cError29, cError30, cError31, cError32, cError33,
                    };

unsigned char    ilc;

unsigned long           ulCCCount;
unsigned long           ulCCCount1;
unsigned long           ulCCCount2;
unsigned long           ulCCCount3;
unsigned long           ulCCCount4;
unsigned long           ulCCCount5;
unsigned long           ulCCCount6;

//---------------------------------------------------------------------------
// function declarations



//---------------------------------------------------------------------------

void    vDisplayArrayElements (FILE * FilePointer, unsigned char * ArrayPointer, unsigned long DisplayLength, unsigned char ArrayType)
{
    unsigned long   i,j;

    if ((ArrayPointer == NULL) || (DisplayLength == 0) || (ArrayType > DOUBLEARRAY))
        return;

    j = 0;
    switch (ArrayType)
    {
    case VOIDARRAY:         // void
        {
            fprintf (FilePointer, "0: ");
            for (i=0; i<DisplayLength; i++)
            {
                fprintf (FilePointer, " %d", (unsigned char) ArrayPointer[i]);
                j++;
                if (j >= 10)
                {
                    j = 0;
                    fprintf (FilePointer, "\n%d: ", i);
                }

            }
        }
        break;

    case CHARARRAY:         // char
        {
            fprintf (FilePointer, "0: ");
            for (i=0; i<DisplayLength; i++)
            {
                fprintf (FilePointer, " %d", (char) ArrayPointer[i]);
                j++;
                if (j >= 10)
                {
                    j = 0;
                    fprintf (FilePointer, "\n%d: ", i);
                }

            }
        }
        break;

    case UNSIGNEDCHARARRAY: // unsigned char
        {
           fprintf (FilePointer, "0: ");
           for (i=0; i<DisplayLength; i++)
           {
                fprintf (FilePointer, " %d", (unsigned char) ArrayPointer[i]);
                j++;
                if (j >= 10)
                {
                    j = 0;
                    fprintf (FilePointer, "\n%d: ", i);
                }

            }
        }
        break;

    case INTARRAY:          // int
        {
            fprintf (FilePointer, "0: ");
            for (i=0; i<DisplayLength; i++)
            {
                fprintf (FilePointer, " %d", (int) ArrayPointer[i]);
                j++;
                if (j >= 10)
                {
                    j = 0;
                    fprintf (FilePointer, "\n%d: ", i);
                }
            }
        }
        break;

    case UNSIGNEDINTARRAY:  // unsigned int
        {
            fprintf (FilePointer, "0: ");
            for (i=0; i<DisplayLength; i++)
            {
                fprintf (FilePointer, " %d", (unsigned int) ArrayPointer[i]);
                j++;
                if (j >= 10)
                {
                    j = 0;
                    fprintf (FilePointer, "\n%d: ", i);
                }
            }
        }
        break;

    case LONGARRAY:         // long
        {
            fprintf (FilePointer, "0: ");
           for (i=0; i<DisplayLength; i++)
           {
                fprintf (FilePointer, " %d", (long) ArrayPointer[i]);
                j++;
                if (j >= 10)
                {
                    j = 0;
                    fprintf (FilePointer, "\n%d: ", i);
                }

            }
        }
        break;

    case UNSIGNEDLONGARRAY: // unsigned long
        {
            fprintf (FilePointer, "0: ");
            for (i=0; i<DisplayLength; i++)
            {
                fprintf (FilePointer, " %d", (unsigned long) ArrayPointer[i]);
                j++;
                if (j >= 10)
                {
                    j = 0;
                    fprintf (FilePointer, "\n%d: ", i);
                }

            }
        }
        break;

    case FLOATARRAY:        // float
        {
            fprintf (FilePointer, "0: ");
            for (i=0; i<DisplayLength; i++)
            {
                fprintf (FilePointer, " %8.4f", (float) ArrayPointer[i]);
                j++;
                if (j >= 10)
                {
                    j = 0;
                    fprintf (FilePointer, "\n%d: ", i);
                }

            }
        }
        break;

    case DOUBLEARRAY:       // double
        {
            fprintf (FilePointer, "0: ");
            for (i=0; i<DisplayLength; i++)
            {
                fprintf (FilePointer, " %8.4f", (double) ArrayPointer[i]);
                j++;
                if (j >= 10)
                {
                    j = 0;
                    fprintf (FilePointer, "\n%d: ", i);
                }

            }
        }
        break;

    default:                // void
        {
            fprintf (FilePointer, "0: ");
            for (i=0; i<DisplayLength; i++)
            {
                fprintf (FilePointer, " %d", (unsigned char) ArrayPointer[i]);
                j++;
                if (j >= 10)
                {
                    j = 0;
                    fprintf (FilePointer, "\n%d: ", i);
                }

            }
        }
        break;
    }       // end of switch (ArrayType)

    fprintf (FilePointer, "\n\n");
}

//---------------------------------------------------------------------------
int iDiscFourierXfm (long idx, int win_flg)
{
    //  read DFT_SIZE values from input_file, starting at idx;
    //  apply proper window function;
    //  take DFT;
    //
    //  Return:    -1 if seek failed
    //              1 if end of file was reached, while trying to read
    //              0 otherwise

    int             read_cnt;
    int             eof_flag;
    unsigned int    si;

    eof_flag = 0;
    if (fseek(FpInputFile, ulHeaderLengthB + (idx>=0 ? idx*sizeof(W2B_INPUT_VAR_TYPE): 0L), SEEK_SET) == 0)
    {
        read_cnt = fread(&datarI[0], sizeof(W2B_INPUT_VAR_TYPE), DFT_SIZE, FpInputFile);

        if (read_cnt != DFT_SIZE)
        {
            eof_flag = 1;
            for (si=read_cnt; si<DFT_SIZE; si++)
                datarI[si]=0.0;
        }

        if (win_flg == 1)
             for (si=0; si<DFT_SIZE; si++)
                fInputReal[si] = ((float) datarI[si]) * wind[si];    // 2002 may06
        else
            for (si=0; si<DFT_SIZE; si++)
                fInputReal[si] = (float) datarI[si];

        //-------------------------------------------------------------------
        if (uiDebugLevel == 25)
        {
            fprintf (FpDebugFile, "iDiscFourierXfm: start fInputReal[]/fOutputReal data:\n\n");
            vDisplayArrayElements (FpDebugFile, (unsigned char *) fInputReal, 8, FLOATARRAY);
            vDisplayArrayElements (FpDebugFile, (unsigned char *) fOutputReal, 8, FLOATARRAY);
        }
        //-------------------------------------------------------------------

        dfft3f (fInputReal, fInputImag, fOutputReal, fOutputImag, snqt, DFT_SIZE-1, 0, 1);

        //-------------------------------------------------------------------
    if (uiDebugLevel == 25)
    {
        fprintf (FpDebugFile, "iDiscFourierXfm: start fInputReal[]/fOutputReal data:\n\n");
        vDisplayArrayElements (FpDebugFile, (unsigned char *) fInputReal, 8, FLOATARRAY);
        vDisplayArrayElements (FpDebugFile, (unsigned char *) fOutputReal, 8, FLOATARRAY);
    }
        //-------------------------------------------------------------------

        return (eof_flag);
    }    // end of successful seek
    else
        return (-1);   // seek failed
}

//---------------------------------------------------------------------------
int iWDiscFourierXfm (long lStartIndex, int iWindowFlag)
{
    // read DFT_SIZE_W values from input_file, starting at lStartIndex
    // apply proper window function
    // take DFT
    //
    //    Return:  -1 if seek failed
    //              1 if end of file was reached, while trying to read
    //              0 otherwise

    int             read_cnt;
    int             eof_flag;
    unsigned int    si;
    float           fmin, fmax;

    eof_flag = 0;
    if (fseek (FpInputFile, ulHeaderLengthB + lStartIndex*sizeof(W2B_INPUT_VAR_TYPE), SEEK_SET) == 0)
    {
        read_cnt = fread (&datarI[0], sizeof(W2B_INPUT_VAR_TYPE), DFT_SIZE_W, FpInputFile);
        if (read_cnt != DFT_SIZE_W)
        {
            eof_flag = 1;
            for (si=read_cnt; si<DFT_SIZE_W; si++)
                datarI[si] = 0.0;
        }

        if (iWindowFlag == 1)
            for (si=0; si<DFT_SIZE_W; si++)
                fInputReal[si]  = ((float) datarI[si]) * windW[si];    // 2002 may06
        else
            for (si=0; si<DFT_SIZE_W; si++)
                fInputReal[si]  =  (float) datarI[si];

        //-------------------------------------------------------------------
        if (uiDebugLevel == 25)
        {
            fprintf (FpDebugFile, "iWDiscFourierXfm: iWindowFlag = %d\n", iWindowFlag);
            fprintf (FpDebugFile, "iWDiscFourierXfm: start fInputReal[]/fOutputReal data:\n\n");
            vDisplayArrayElements (FpDebugFile, (unsigned char *) fInputReal,  8, FLOATARRAY);
            vDisplayArrayElements (FpDebugFile, (unsigned char *) fOutputReal, 8, FLOATARRAY);
        }
        //-------------------------------------------------------------------

        dfft3f (fInputReal, fInputImag, fOutputReal, fOutputImag, snqtW, DFT_SIZE_W-1, 0, 1);

        //-------------------------------------------------------------------
        if (uiDebugLevel == 25)
        {
            fprintf (FpDebugFile, "iWDiscFourierXfm: end fInputReal[]/fOutputReal data:\n\n");
            vDisplayArrayElements (FpDebugFile, (unsigned char *) fInputReal, 8, FLOATARRAY);
            vDisplayArrayElements (FpDebugFile, (unsigned char *) fOutputReal, 8, FLOATARRAY);

            fmin = fInputReal[0];
            fmax = fInputReal[0];

              for (si=1; si<DFT_SIZE_W; si++)
            {
                if (fInputReal[si] < fmin)
                    fmin = fInputReal[si];
                else if (fInputReal[si] > fmax)
                    fmax = fInputReal[si];
            }

            fprintf (FpDebugFile,"iWDiscFourierXfm: min= %10.3f max= %10.3f\n\n", fmin, fmax);
        }
        //-------------------------------------------------------------------

        return (eof_flag);
    }    // end of successful seek

    else
        return (-1);  // seek failed
}

//---------------------------------------------------------------------------
void    set_linear_magnitude (float * fInputCosine, float * fInputSine, float * fMagnitude, float * fArgument, unsigned int uiCount)
{
    // function to calculate linear magnitude of complex sequence
    //
    //      input parameter 'fArgument' is not used in this function

    unsigned int    i;
    double          dtmp1, dtmp2, dtmp3;

    for (i=0; i<uiCount; i++)
    {
        dtmp1           = (double) fInputCosine[i];
        dtmp2           = (double) fInputSine[i];
        dtmp3           = sqrt (dtmp1*dtmp1 + dtmp2*dtmp2);
        dtmp3           = (dtmp3>1e-10?dtmp3:1e-3);

        fMagnitude[i]   = (float) dtmp3;
    }
}


//---------------------------------------------------------------------------
int    iDecode (int *codeword, int rho, int *erase)
{
    // 'codeword' points to the codeword to be decoded.  This is  in the
    //            form [r r r r ... i i i i i i], where the r's are the
    //            redundancy symbols, and the i's are the information symbols,
    //            both with least significant symbols first.
    //
    // 'rho'      is the number of erasures
    // 'erase'    points to an array of erasure indexes.  Each index is
    //            in the range 0 through MAX_GF_EL_1-1
    //
    //  returns    a positive number indicating how many symbols were corrected
    //
    //       0 if no symbols were in error
    //      -1 if any codeword symbols are illegal
    //      -2 if rho > 2*t (too many erasures)
    //      -3 if any erasure indexes are out of bounds
    //
    //      if return value >= 0, then *codeword contains the corrected values.

    int        si, di;
    int        V[MAX_GF_EL_1], S[MAX_GF_EL_1], lam[MAX_GF_EL_1], B[MAX_GF_EL_1], T[MAX_GF_EL_1],
               del[NUM_GF_EL_1], C[MAX_GF_EL_1], c[MAX_GF_EL_1];
    int        sh[MAX_GF_EL_1];
    int        r, L, tmpi, tmpj, tmpk, texp, ec;

    for (si=0; si<MAX_GF_EL_1; si++)
        if ((codeword[si] < 0) || (codeword[si] > MAX_GF_EL_1))
            return -1;

    if (rho > (2*t1))
        return -2;

    if (rho > 0)
    {
        for (si=0; si< rho; si++)
        if ((erase[si] < 0) || (erase[si] > (MAX_GF_EL_1-1)))
            return -3;
    }

    L      = 0;
    r      = 0;
    B[0]   = 1;
    lam[0] = 1;
    for (si=1; si<MAX_GF_EL_1; si++)
        {B[si]=0; lam[si]=0;}

    // Transform received codeword
    for (di=0; di<MAX_GF_EL_1; di++)
    {
        tmpi = 0;
        for (si=0; si<MAX_GF_EL_1; si++)
            tmpi += expa1[(si*di)%MAX_GF_EL_1] * codeword[si];
        V[di] = tmpi%NUM_GF_EL_1;
        S[di] = V[di];
    }

    // note that while doing transforms S indexes range from 0 through MAX_GF_EL_1-1, but
    // when revising values of S the index ranges from 1 to MAX_GF_EL_1.  This is
    // resolved by using 0 through MAX_GF_EL_1-1 while transforming, and using
    // 1 through MAX_GF_EL_1 (mod MAX_GF_EL_1) when revising values of S.

    while (r<rho)
    {
        r++;
        {
            for (si=MAX_GF_EL_1-1; si>0; si--)
                sh[si] = (expa1[erase[r-1]]*lam[si-1])%NUM_GF_EL_1;
            sh[0] = 0;
            for (si=0; si<MAX_GF_EL_1; si++)
                {lam[si]= (lam[si] - sh[si])%NUM_GF_EL_1; B[si]= lam[si];}
            L++;
        }
    } // end of (r<rho)

    while (r < 2*t1)
    {
        r++;
        // calculate del[r]
        tmpi = 0;
        for (si=0; si<=(r<MAX_GF_EL_1 ? r: MAX_GF_EL_1-1); si++)
            tmpi += lam[si]*S[(r-si)%MAX_GF_EL_1];
        del[r%MAX_GF_EL_1] = tmpi%NUM_GF_EL_1;
        while(del[r%MAX_GF_EL_1] < 0)
            del[r%MAX_GF_EL_1]+=NUM_GF_EL_1;
        if (del[r] == 0)
        {
            for (si=MAX_GF_EL_1-1; si>0; si--)
            B[si] = B[si-1];
            B[0]  = 0;
        } // end of del[r]==0
        else
        {
            for (si=1; si<MAX_GF_EL_1; si++)
                sh[si] = (del[r%MAX_GF_EL_1]*B[si-1])%NUM_GF_EL_1;
            sh[0] = 0;
            for (si=0; si<MAX_GF_EL_1; si++)
                T[si] = (lam[si]-sh[si])%NUM_GF_EL_1;
            if (2*L > r+rho-1)
            {
                for (si=0; si<MAX_GF_EL_1; si++)
                    lam[si]= T[si];
                for (si=MAX_GF_EL_1-1; si>0; si--)
                    B[si] = B[si-1];
                B[0] = 0;
            } // end of no change in L
            else
            {
                L    = r - L + rho;
                tmpi = inve1[del[r]];
                for (si=0; si<MAX_GF_EL_1; si++)
                    B[si] = (tmpi*lam[si])%NUM_GF_EL_1;
                for (si=0; si<MAX_GF_EL_1; si++)
                    lam[si] = T[si];
            } // end of increase in L
        } // end of del[r]!=0
    } // end of (r <= 2*t1)

    while(r<MAX_GF_EL_1)
    {
        r++;
        // calculate del[r]
        tmpi = 0;
        for (si=0; si<=(r<MAX_GF_EL_1 ? r: MAX_GF_EL_1-1); si++)
            tmpi += lam[si]*S[(r-si)%MAX_GF_EL_1];
        del[r%MAX_GF_EL_1] = tmpi%NUM_GF_EL_1;
        while(del[r%MAX_GF_EL_1] < 0)
            del[r%MAX_GF_EL_1]+=NUM_GF_EL_1;
        S[r%MAX_GF_EL_1] = (S[r%MAX_GF_EL_1]-del[r%MAX_GF_EL_1])%NUM_GF_EL_1;
    } // end of (r<=MAX_GF_EL_1)

    // correct transformed codeword
    for (si=0; si<MAX_GF_EL_1; si++)
        C[si] = (V[si]-S[si])%NUM_GF_EL_1;

    // inverse transform corrected codeword
    tmpi = inve1[MAX_GF_EL_1];
    for (si=0; si<MAX_GF_EL_1; si++)
    {
        tmpj = 0;
        for (di=0; di<MAX_GF_EL_1; di++)
        {           // tmpk = a^(-si*di)
            texp = (si*di)%MAX_GF_EL_1;
            if (texp == 0)
                tmpk = 1;
            else
                tmpk = inva1[texp] ;
             tmpj += (tmpk*C[di])%NUM_GF_EL_1;
        } // end of inner inverse transform loop
        c[si] = (tmpi*tmpj)%NUM_GF_EL_1;
        while(c[si] < 0)
            c[si] += NUM_GF_EL_1;
    } // end of inverse transform

    // correct and count wrong symbols
    ec = 0;
    for (si=0; si<MAX_GF_EL_1; si++)
    {
        if (c[si] != codeword[si])
            {ec++; codeword[si]=c[si];}
    }

    return ec;
}

//---------------------------------------------------------------------------
int    lm_match3(float *lmfreq, int *lmidx, int count, int *lmgoodi, double fdthld)
{
    //  check for spacing of lmfreq[] compared to ideal values
    //  return number of "good" cases found, or -1 if an error is detected

    float   md, nd;
    int     si, di, max_cnt, init_flg, gcnt, lc[num_sc_l];

    init_flg = 1;
    max_cnt  = 0;
    for (si=0; si<count; si++)
        lc[si]=0;
    for (di=0; di<count; di++)
    {
        for (si=0; si<count; si++)
        {
            if (si==di)
                continue;
            md  = lmfreq[di] - lmfreq[si];
            nd = pm_nom_lt[lmidx[di]] - pm_nom_lt[lmidx[si]];
            if (fabs((double)(md/nd -1.0)) < fdthld)
            {
                lc[di]++;
                if (init_flg==1)
                    {init_flg=0; max_cnt=1;}
                else
                    {if (lc[di]>max_cnt) max_cnt= lc[di];}
            } // end of match found
        } // end of for si
    } // end of for di

    gcnt = 0;
    for (si=0; si<count; si++)
    {
        if (lc[si]>= (count-3))
            lmgoodi[gcnt++] = lmidx[si];
    }

    if (gcnt>= nsc_min)
        return(gcnt);
    else
        return(-1);    // error case
}

//---------------------------------------------------------------------------
int    iFindLocalMax (float *mag, float *lmfreq, float *lmqual, int *lmidx, unsigned int *bm)
{
    //  find local max values near indexes listed in exidx[],
    //  return number of local max values found

    int     count, max_mask, lm_map;
    int     si, di, bi, ni, ai;
    float   acumn, acumd, blm, tmpa, dummyf;

    max_mask = 0x0001;
    count    = 0;
    lm_map   = 0;
    for (si=0; si< num_sc_l; si++)
    {
        ni = exidx[si];
        for (di= -ossr2; di<= ossr2; di++)
        {
            if (di== -ossr2)
                {blm= mag[ni+di]; bi=di+ni;}
            if ((mag[ni+di-2] < mag[ni+di-1]) &&
                (mag[ni+di-1] < mag[ni+di])   &&
                (mag[ni+di]   > mag[ni+di+1]) &&
                (mag[ni+di+1] > mag[ni+di+2]))
            {
                if (mag[ni+di] > blm)
                    {blm= mag[ni+di]; bi=di+ni;}
                lm_map |= max_mask;
            } // end of local max found
        } // end of for di

        if ((lm_map & max_mask) == max_mask)
        {                                // calculate weighted average for this local max
            lmidx[count] = si;
            acumn          = 0.;
            acumd          = 0.;
            for (ai=-1; ai<2; ai++)
            {
                tmpa   = mag[bi+ai];
                dummyf = (float)(bi+ai)*(float)(f_stepd);
                acumn += tmpa*dummyf;
                acumd += tmpa;
            }

            lmfreq[count] = acumn/acumd;
            acumd           = 0.;                    // calculate quality number for this local max
            for (ai=-1; ai<2; ai++)
            {
                acumd += mag[bi+ gap_offi+ai] + mag[bi - gap_offi + ai];
            }
            acumd         = (acumd > 1e-3 ? acumd : 1e-3);
            lmqual[count] = mag[bi]*6.0/acumd;
            count++;
        }   // end of bit set for local max found
        max_mask = max_mask << 1;
    } // end of for si

    bm[0] = lm_map;

    return(count);
}

//---------------------------------------------------------------------------
int    iWFindLocalMax (float *mag, float *lmfreq, float *lmqual, int *lmidx, unsigned int *bm)
{
    //  find local max values near indexes listed in exidx[],
    //  return number of local max values found

    int     count, max_mask, lm_map;
    int     si, di, bi, ni, ai;
    float   acumn, acumd, blm, tmpa, dummyf;

    max_mask = 0x0001;
    count    = 0;
    lm_map   = 0;
    for (si=0; si< num_sc_l; si++)
    {
        ni=exidxW[si];
        for (di= -2; di<= 2; di++)
        {
            if (di== -2)
                {blm= mag[ni+di]; bi=di+ni;}
            if ((mag[ni+di-1] < mag[ni+di]) && (mag[ni+di] > mag[ni+di+1]))
            {
                if (mag[ni+di] > blm)
                    {blm= mag[ni+di]; bi=di+ni;}
                lm_map |= max_mask;
            } // end of local max found
        } // end of for di
        if ((lm_map & max_mask) == max_mask)
        {
            lmidx[count] = si;
            // calculate weighted average for this local max
            acumn = 0.;
            acumd = 0.;
            for (ai=-1; ai<2; ai++)
            {
                tmpa   = mag[bi+ai];
                dummyf = (float)(bi+ai)*(float)(f_stepdW);
                acumn += tmpa*dummyf;
                acumd += tmpa;
            }
            lmfreq[count] = acumn/acumd;
            // calculate quality number for this local max
            acumd = 0.;
            for (ai=-1; ai<2; ai++)
            {
                acumd += mag[bi+ gap_offi/4 +ai] + mag[bi- gap_offi/4 +ai];
            }
            acumd         = (acumd > 1e-3 ? acumd : 1e-3);
            lmqual[count] = mag[bi]*6.0/acumd;
            count++;
        }// end of bit set for local max found
        max_mask = max_mask << 1;
    } // end of for si

    bm[0] = lm_map;
    return(count);
}


//---------------------------------------------------------------------------
float    fCrossCorrelation (float * fpPatternArray, float * fpDataArray, unsigned int uiPatternLength, unsigned int uiCount, unsigned int * uipCountIndex)
{

    // Floating point cross correlation calculation
    //
    //      fpPatternArray  pointer to pattern array
    //      fpDataArray     pointer to data array
    //      uiPatternLength length of pattern array
    //      uiCount         number of offsets into data array at which to calculate correlation coefficient
    //      uipCountIndex   points to where "uiCount" value for greatest correlation coefficient value is to be stored
    //
    //      returns         correlation coefficient between pattern[] and data[uipCountIndex] (which is greatest value found)

    double          sump, sump2, sumd, sumd2, sumpd;
    double          n, dp, dd, patd, tmpd, ccmax;
    unsigned int    si, di, cm;

    ulCCCount++;

    n     = (double) uiPatternLength;
    sump  = 0.0;
    sump2 = 0.0;
    for (si=0; si<uiPatternLength; si++)
    {
        dp     = fpPatternArray[si];    /*??*/
        sump  += dp;
        sump2 += dp * dp;
    }

    patd = n * sump2 - sump * sump;

    for (di=0; di<uiCount; di++)
    {
        sumd  = 0.0;
        sumd2 = 0.0;
        sumpd = 0.0;
        for (si=0; si<uiPatternLength; si++)
        {
            dp     = fpPatternArray[si];
            dd     = fpDataArray[si+di];       // how big is 'data' array?
            sumd  += dd;
            sumd2 += dd * dd;
            sumpd += dd * dp;
        }

        tmpd = ((n * sumpd - sump * sumd) / sqrt (patd * (n * sumd2 - sumd * sumd)));

        if (di==0)
        {
            cm    = di;
            ccmax = tmpd;
        }
        else
        {
            if (tmpd > ccmax)
            {
                cm    = di;
                ccmax = tmpd;
            }
        }
    } // end of for di

    *uipCountIndex = cm;
    return (float) ccmax;
}

//---------------------------------------------------------------------------
int    iCalcBestFit (float *lmfreq, int *lmidx, int count, float *slope, float *intercept)
{
    //  calculate least squares best fit between contents of lmfreq[] and
    //  the count relevant entries in pm_nom_lt[]
    //
    //  return 0 on success, -1 on error

    int     di;
    float   sumxy, sumx, sumy, x2sum, cx;

    // least squares fit to coarse tuning errors
    sumx  = 0.0;
    sumy  = 0.0;
    sumxy = 0.0;
    x2sum = 0.0;
    for (di=0; di<count; di++)
    {
        cx     = pm_nom_lt[lmidx[di]];
        sumx  += cx;
        x2sum += cx*cx;
        sumxy += cx*lmfreq[di];
        sumy  += lmfreq[di];
    }

    *slope     = ((float)(count)*sumxy - sumx*sumy)/(count*x2sum - sumx*sumx);
    *intercept = sumy/(float)(count) - *slope*sumx/(float)(count);

    return(0);
}

//---------------------------------------------------------------------------
int    iFindLocalMaxAmp (float *data, int ossi, int *lm_off, float *lm_value)
{
    //  Search no more than ossi locations away from the starting location
    //  for a local maximum in the cross-correlation between the data[] values
    //  and the am0_7ref[] pattern.
    //
    //  Return 1 and set lm_off[0] and lm_value[] if local max found,
    //  otherwise return 0

    float   cval, pval;
    int     idx, istep, sii, cmx;

    pval = fCrossCorrelation (&am_ref[0], &data[0], am_ref_len, 1, &cmx);
    ulCCCount3++;
    cval = fCrossCorrelation (&am_ref[0], &data[1], am_ref_len, 1, &cmx);
    ulCCCount4++;

    if (pval == cval)
        return(0); // fail on adjacent equal values

    if (cval > pval)
    {
        istep = 1;
        idx   = 1;
    }
    else
    {
        istep = -1;
        idx   = 0;
    }

    for (sii=0; sii<ossi; sii++)
    {
        pval  = cval;
        idx  += istep;
        cval  = fCrossCorrelation (&am_ref[0], &data[idx], am_ref_len, 1, &cmx);
        ulCCCount5++;

        if (cval == pval)
            return(0);
        if (cval < pval)
        {
           lm_off[0]   = idx - istep;
           lm_value[0] = pval;
           return(1);
        }
    }// end of for sii

    return(0);
}

//---------------------------------------------------------------------------
void    error_msg_to_stderr (int ern)
{
    fprintf(stderr,"%s\n", erexmsg[ern]);
    ilc='*';
}

//---------------------------------------------------------------------------
int main (int argc, char *argv[], char *env[])
{
    int             ERN, err, j, ki, eci;
    unsigned int    err_loc = 0;
    int             si, di;
    int             eof_flag, win_flg;

    struct          stat    statb, statir;

    int             statr, statirr;
    unsigned char   lss;

    double          t_step1st;
    double          dummyd1;
    float           fdummy_1;
    float           r53 [3];

    int             mi, lsi, lo, fi;
    float           cx, sumx, x2sum, sumxy, sumy, sumx2;
    double          pm_ft_ctr[NUM_TOTAL_SC];
    float           f_slope, f_intercept;
    double          clk_period, ca, astep, ang_ref, ang_off;
    double          bias[num_sc_l], pv[num_sc_l];

    int             end_of_file;
    int             val_to_read, dum_len, val_to_write;
    size_t          val_read, val_written;

    unsigned long   ulFileLength, ulFIndeXB, ulNumberOfElements, ulFileLength_dmd, ulNumberOfElements_dmd;
    unsigned long   ulInputElRead;

    unsigned char   scratch1[] = "w2bscr1.tmp";
    unsigned char   scratch2[] = "w2bscr2.tmp";

    float         * fbufin;
    float         * fbufot;

    int             delay, ns, dmd_cnt;
    double          NS, acum0d;
    float           acum0, acum1, dummyf1, msenom;

    int             ioffW, dmd_lidx;
    float           cppblk, cppblk_fr, cppblk_ft;
    int             num_frsq, dmd_fidx, fridxoff[NUM_TOTAL_SC], bfridx;
    int           * fridxadj[NUM_TOTAL_SC], bk;
    int             fridxoffsr[NUM_TOTAL_SC], slope_valid[NUM_TOTAL_SC], scv;
    double          frsq_mmse[NUM_TOTAL_SC], bfrsq_mmse;

    double          dcidx_step[NUM_TOTAL_SC];
    int             cbidx, cbidxm, eracnt1, scc, scca, eracnt1s;
    int             eracnt1ss, eracnt1sss, eracnt1ssss;
    int             erasi1[2][SYMBOLS_PER_CODEBLOCK], eccnt1[2];
    int             erasi1s[2][SYMBOLS_PER_CODEBLOCK], hec, sec, erag;
    int             erasi1ss[2][SYMBOLS_PER_CODEBLOCK], ssec;
    int             erasi1sss[2][SYMBOLS_PER_CODEBLOCK], sssec, ssssec;
    float           distg, distg_last;
    int             sccx[NUM_TOTAL_SC];

    unsigned char   fotbuf [SYMBOLS_PER_CODEBLOCK * 2];         // buffer for output file
    unsigned char   fnbuf[MAX_FILENAME_LENGTH+1];               // buffer for file name

    SYM_TYPE_SC  *  cbmerged, cbmd [SYMBOLS_PER_CODEBLOCK * 2], cbmr[SYMBOLS_PER_CODEBLOCK * 2];
    int             tmpi, fbtow, fbw, fwr, fnlen, file_lengthb;
    float           c2phbuf[NUM_TOTAL_SC];

    int             bfridx0;

    int             default_mode;
    int             fmatch, fmatchi[num_sc_l], ji, jfm;
    float           as_off;
    float         * frsqmse1;
    int             frsqmse1i;

    float           frsqmseave [NUM_TOTAL_SC];
    float           frsqmsesd [NUM_TOTAL_SC];

    unsigned char * fridxbetter [NUM_TOTAL_SC];
    float           cppblk_slope [num_sc_l];
    float           cppblk_off [num_sc_l];
    int             era_rsv [num_ec_lvls] = {8, 12, 20, 48};
    int             bad_blks;
    int             filename_OK_flag;
    double          Ampd;
    double          NORM;
    int             demod_lcnt;

    unsigned int    ob_cnt, ob_mask, bec_mask;
    int             di_last, scc_last;

    double          sym_pos [NUM_TOTAL_SC] [MAX_GF_EL_1+frame_len], ar_step0;
    unsigned char   ma1 [NUM_TOTAL_SC] [MAX_GF_EL_1+frame_len];
    int             sas, spni, lm_off_amp, ar_state, aridx;
    int             arldx, arzdx, arudx, nzidxl, nzidxr, ii0;
    float           lm_value_amp;

    unsigned long   ulLoopCount;

    //-----------------------------------------------------------------------
    //  main () begins here

    //  arv[0]
    //  arv[1] is input file name
    //  arv[2] is debug level
    //  arv[3] is mode
    //  arv[4] is clock frequency

    //-----------------------------------------------------------------------
    // print sign on message
    fprintf (stdout, "\n%s - %s.%s\n\n", cSignonMessage, cVersionMajor, cVersionMinor);

    //-----------------------------------------------------------------------
    // process command line parameters
    if (argc < 2)                                   // enough arguments?
        {fprintf (stderr,"%s\n", cUsage);  exit(-1);}

    //-----------------------------------------------------------------------
    // argv[1] is input file
    statr  = stat (argv[1], &statb);
    if (statr != 0)
        {ERN=6; goto exit_j;}

    ulFileLength        = (unsigned long) (statb.st_size);
    ulNumberOfElements  = (ulFileLength - ulHeaderLengthB) / (sizeof(W2B_INPUT_VAR_TYPE));

    //-----------------------------------------------------------------------
    // argv[2] is debug level
    if (argc > 2)
    {
        uiDebugLevel = (unsigned int) atoi (argv[2]);
    }
    else
        uiDebugLevel = DEFAULTDEBUGLEVEL;

    if (uiDebugLevel != DEFAULTDEBUGLEVEL)
    {
        fprintf (stdout, "  Using Debug Level = %d\n\n", uiDebugLevel);
    }

    //-----------------------------------------------------------------------
    // argv[3] is mode
    if (argc > 3)
        default_mode=atoi(argv[3]);
    else
        default_mode=0;

    //-----------------------------------------------------------------------
    // argv[4] is clock frequency
    if (argc > 4)
    {
        clk_period = 1.0 / atof (argv[4]);
        if ((clk_period > 500e-6) || (clk_period < 10e-6))
            {ERN=31; goto exit_j;}
    }
    else                                            // default frequency
        clk_period = 1.0 / clk_freq_nom;            // 11025.0
    t_step = (float) clk_period;

    //-----------------------------------------------------------------------
    // initialize stuff
    init_code2();                                   // finish initialization of distance threshold arrays

    //-----------------------------------------------------------------------
    // initialize debug stuff
    ulCCCount        = 0;
    ulCCCount1       = 0;
    ulCCCount2       = 0;
    ulCCCount3       = 0;
    ulCCCount4       = 0;
    ulCCCount5       = 0;
    ulCCCount6       = 0;
    ulLoopCount      = 0;

    //-----------------------------------------------------------------------
    // open input file
    FpInputFile = fopen(argv[1], "rb");              // open input file
    if (FpInputFile == NULL)
        {fprintf(stderr,"Could not open file %s\n", argv[1]); exit(-7);}

    // validate input file name/type
    if (strstr(argv[1],".wav") != NULL)             // check input filename extension
        ulHeaderLengthB = wavhdrlenb;
    else if (strstr(argv[1],".sw") != NULL)
        ulHeaderLengthB = 0L;
    else
    {
        fprintf (stderr,".wav and .sw are the only file extensions supported by this program\n");
        fclose (FpInputFile);
        exit (-6);
    }

    cpOutputFilename = argv[1];

    //-----------------------------------------------------------------------
    if (uiDebugLevel == 25)                // 2002 apr19
    {
        remove (cDebugFilename);                            // delete previous debug file
        FpDebugFile = NULL;
        FpDebugFile = fopen (cDebugFilename, "wt");           // open file for debug data // 2002 apr19
        if (FpDebugFile == NULL)
        {
            fprintf (stderr,"Could not open file w2b_dbg.txt\n");
            fclose (FpDebugFile);
            exit(-8);
        }
    }
    //-----------------------------------------------------------------------

    //-----------------------------------------------------------------------
    // init stuff for input file processing

    ulSourceIndex   = 0L;
    lss             = 0x00;
    sstepW          = 750L;
    ERN             = 3;

    t_step1st       = (double) t_step;
    f_stepd         = init_dft3 (DFT_SIZE,   t_step, snqt);
    f_stepdW        = init_dft3 (DFT_SIZE_W, t_step, snqtW);

    //-----------------------------------------------------------------------
    if (uiDebugLevel == 25)
    {
        fprintf (FpDebugFile, "main: start of snqtW data:\n\n");
        vDisplayArrayElements (FpDebugFile, (unsigned char *) snqtW, 8, FLOATARRAY);
    }
    //-----------------------------------------------------------------------

    //-----------------------------------------------------------------------
    // init window arrays
    dummyd1   = PI2 / (double) (DFT_SIZE-1);
    for (si=0; si<DFT_SIZE; si++)
        wind[si]  = (float) (0.54 - 0.46 * cos(dummyd1 * (double) si));

    dummyd1   = PI2 / (double) (DFT_SIZE_W-1);
    for (si=0; si< DFT_SIZE_W; si++)
        windW[si] = (float) (0.54 - 0.46 * cos(dummyd1 * (double) si));

    win_flg = 1;
    fprintf (stdout,"For data file: %s\n", argv[1]);
    if (win_flg == 1)
        fprintf(stdout,"\tusing Hamming window and %s program\n", argv[0]);
//    else
//        fprintf(stdout,"\tusing Rectangular window and %s program\n", argv[0]);

    //-----------------------------------------------------------------------
    // find leader, restrict search to first half of file
    lsi             = 2;
    lead_x[0].state = 0;
    lead_x[0].mode  = -1;           // an illegal value
    while ((ulSourceIndex < ulNumberOfElements/2) && (lss == 0x00))
    {
        eof_flag = iWDiscFourierXfm (ulSourceIndex, win_flg);
        if (eof_flag < 0)
            {ERN=4; goto exit_j;}    // seek failed

        //-------------------------------------------------------------------
        if (uiDebugLevel == 25)
        {
            ulLoopCount++;
            fprintf (FpDebugFile, "\n//---------------------------------------------------------------------\n// ulLoopCount = %d\n\n", ulLoopCount);
            fprintf (FpDebugFile, "main: start of output data from FFT, input data to set_linear_magnitude:\n\n");
            fprintf (FpDebugFile, "    real       imaginary\n");
            for (j=0; j<32; j++)
                fprintf (FpDebugFile, " %8.1f   %8.1f\n", fOutputReal[j], fOutputImag[j]);
            fprintf (FpDebugFile,"\n\n");
        }
        //-------------------------------------------------------------------

        set_linear_magnitude (fOutputReal, fOutputImag, &mag[lsi][0], &mag[lsi][0], DFT_SIZE_W);

        lead_x[lsi].xcv = fCrossCorrelation (&trail_spec_nom[0], &mag[lsi][lsi_startW], trail_spec_len_nom, lsi_countW, &lead_x[lsi].offi);
        ulCCCount1++;

        mi=0;
        while ((mi<num_modes_l) && (lss==0x00))
        {
            //---------------------------------------------------------------
            if (uiDebugLevel == 25)            // 2002 apr19
            {
                fprintf(FpDebugFile,"main: ulSourceIndex= %8ld  lead_x[lsi].xcv= %8.5f  lead_x[lsi].offi= %8d\n\n", ulSourceIndex, lead_x[lsi].xcv, lead_x[lsi].offi); // 2002 apr19
            }
            //---------------------------------------------------------------
            if (lead_x[lsi].xcv > xcthld)
            {
                as_off = (lead_x[lsi].offi + lsi_startW - lsi_centerW)*(float)(f_stepdW);
                for (fi=0; fi< num_sc_l; fi++)
                    exidx[fi]= (int) (floor((double)((pm_nom_lt[fi] + as_off)/(float)(f_stepd)+ 0.5)));

                eof_flag = iDiscFourierXfm (ulSourceIndex + DFT_SIZE_W/2 - DFT_SIZE/2, win_flg);
                if (eof_flag < 0)
                    {ERN=4; goto exit_j;}    // seek failed
                set_linear_magnitude (fOutputReal, fOutputImag, &mag[0][0], &mag[0][0], DFT_SIZE);
                lead_x[0].nlm = iFindLocalMax(&mag[0][0], &lead_x[0].lmfreq[0], &lead_x[0].lmqual[0], &lead_x[0].lmidx[0], &lead_x[0].lmbm);

                //-----------------------------------------------------------
                if (uiDebugLevel == 25)            // 2002 apr19
                {
                    fprintf(FpDebugFile,"main: as_off= %9.4f  lead_x[0].nlm= %2d\n", as_off, lead_x[0].nlm);
                    fprintf(FpDebugFile,"main: exidx[]=  ");
                    for (fi=0; fi< num_sc_l; fi++)
                        fprintf(FpDebugFile," %4d   ", exidx[fi]);
                    fprintf(FpDebugFile,"\n");
                    fprintf(FpDebugFile,"main: lmfreq[]= ");
                    for (fi=0; fi< num_sc_l; fi++)
                        fprintf(FpDebugFile," %7.2f", lead_x[0].lmfreq[fi]);
                    fprintf(FpDebugFile,"\n");
                    fprintf(FpDebugFile,"main: lmidx[]=      ");
                    for (fi=0; fi< num_sc_l; fi++)
                        fprintf(FpDebugFile," %4d", lead_x[0].lmidx[fi]);
                    fprintf(FpDebugFile,"\n");
                }
                //-----------------------------------------------------------

                if (lead_x[0].nlm >= nsc_min)
                {
                    fmatch  = lm_match3(&lead_x[0].lmfreq[0], &lead_x[0].lmidx[0], lead_x[0].nlm, &fmatchi[0], 2.0*fdthld_nom);
                    dummyf1 = 0.0;
                    for (ji=0; ji< lead_x[0].nlm; ji++)
                        dummyf1 += lead_x[0].lmqual[ji];
                    r53[lsi]= (float)(fmatch)*dummyf1/(float)(lead_x[0].nlm);

                    //-------------------------------------------------------
                    if (uiDebugLevel == 25)            // 2002 apr19
                    {
                        fprintf(FpDebugFile,"main: fmatch= %2d  r53[lsi]= %8.3f\n", fmatch, r53[lsi]);
                        fprintf(FpDebugFile,"main: fmatchi[]=  ");
                        for (fi=0; fi< num_sc_l; fi++)
                            fprintf(FpDebugFile," %4d", fmatchi[fi]);
                        fprintf(FpDebugFile,"\n");
                    }
                    //-------------------------------------------------------

                    if (fmatch >= nsc_min)
                    {jfm=0;
                     for (ji=0; ji<lead_x[0].nlm; ji++)
                     {if(lead_x[0].lmidx[ji] == fmatchi[jfm])
                      lead_x[0].lmfreq[jfm++]= lead_x[0].lmfreq[ji];
                     }
                     switch(lead_x[0].state)
                        {
                        case 0:
                            {
                                err = iCalcBestFit(&lead_x[0].lmfreq[0], &fmatchi[0], fmatch, &lead_x[0].slope, &lead_x[0].off_hz);
                                if (err <0)
                                    {err_loc=1; ERN=10; goto exit_j;}
                                lead_x[0].state   = 1;
                                lead_x[0].sidx    = ulSourceIndex+DFT_SIZE_W/2 -DFT_SIZE/2;
                                lead_x[0].r53     = r53[lsi];
                                lead_x[0].mode    = mi;
                                lead_x[0].lmcount = fmatch;
                                for (ji=0; ji<fmatch; ji++)
                                    lead_x[0].lmgoodi[ji]= fmatchi[ji];
                            }
                        break;

                        case 1:
                            if (r53[lsi] > lead_x[0].r53)
                            {
                                err = iCalcBestFit(&lead_x[0].lmfreq[0], &fmatchi[0], fmatch, &lead_x[0].slope, &lead_x[0].off_hz);
                                if (err <0)
                                    {err_loc=1; ERN=10; goto exit_j;}
                                lead_x[0].sidx    = ulSourceIndex+DFT_SIZE_W/2 - DFT_SIZE/2;
                                lead_x[0].r53     = r53[lsi];
                                lead_x[0].mode    = mi;
                                lead_x[0].lmcount = fmatch;
                                for (ji=0; ji<fmatch; ji++)
                                    lead_x[0].lmgoodi[ji]= fmatchi[ji];
                            }
                            else
                            {
                                lead_x[0].state=2;
                            }
                        break;

                        case 2:
                            if (r53[lsi] > lead_x[0].r53)
                            {
                                err = iCalcBestFit(&lead_x[0].lmfreq[0], &fmatchi[0], fmatch, &lead_x[0].slope, &lead_x[0].off_hz);
                                if (err <0)
                                    {err_loc=2; ERN=10; goto exit_j;}
                                lead_x[0].sidx    = ulSourceIndex+DFT_SIZE_W/2 -DFT_SIZE/2;
                                lead_x[0].r53     = r53[lsi];
                                lead_x[0].mode    = mi;
                                lead_x[0].lmcount = fmatch;
                                for (ji=0; ji<fmatch; ji++)
                                    lead_x[0].lmgoodi[ji]= fmatchi[ji];
                           }
                           else
                           {
                                lead_x[0].state  = 3;
                                lss             |= l1_FOUND ;
                           }
                        break;

                        } // end of switch lead_dat.state
                    }  // end of fmatch large enough
                }    // end of at least nsc_min loacl maximums present
            } // end of cross-corelation value large enough
            else
            {
                switch(lead_x[0].state)
                {
                case 1:
                case 2:
                    lead_x[0].state  = 3;
                    lss             |= l1_FOUND ;
                } // end of switch
            } // end of cross-correlation value below threshold
            mi++;
        } // end of while mi < num_modes_l
        ulSourceIndex += sstepW;
    }    // end of while (looking for leader)

    if (lead_x[0].slope > slope_max)
        lead_x[0].slope = slope_max;
    else
    {
        if (lead_x[0].slope < slope_min)
            lead_x[0].slope= slope_min;
    }
    if(lead_x[0].off_hz > off_hz_max)
        lead_x[0].off_hz = off_hz_max;
    else
    {
        if (lead_x[0].off_hz < off_hz_min)
            lead_x[0].off_hz = off_hz_min;
    }

    //-----------------------------------------------------------------------
    if (uiDebugLevel>10)
    {
        // for now, print results to stdout

        fprintf(stdout,"\nSample index  ---------- leader --------------------\n");
        fprintf(stdout,"              r53 val    offset (Hz)     freq. scale   mode index\n");
        fprintf(stdout,"%8ld     %7.4f     %7.2f           %9.6f      %2d\n",
                lead_x[0].sidx, lead_x[0].r53, lead_x[0].off_hz, lead_x[0].slope, lead_x[0].mode);
    }
    //-----------------------------------------------------------------------

    //-----------------------------------------------------------------------
    if (uiDebugLevel==25)
    {
        fclose (FpDebugFile);
    }
    //-----------------------------------------------------------------------

    if(lss== 0x00)
        {err_loc=3; ERN=10; goto exit_j;}
    for (si=0; si< num_sc_l; si++)
        exidxW[si] = (int)(floor((double)((pm_nom_lt[si] + lead_x[0].off_hz)/(float)(f_stepdW)+ 0.5)));

    // find trailer, restrict search to last half of file
    lead_x[1].state = 0;
    lead_x[1].mode  = -1 ;          // an illegal value
    ulSourceIndex   = ulNumberOfElements - DFT_SIZE_W;
    ioffW           = (int) ((double)(lead_x[0].off_hz)/f_stepdW);
    while((ulSourceIndex > ulNumberOfElements/2) && (lss==0x01))
    {
        eof_flag = iWDiscFourierXfm (ulSourceIndex, win_flg);
        if (eof_flag<0)
            {ERN=4; goto exit_j;}    // seek failed
        set_linear_magnitude (fOutputReal, fOutputImag, &mag[1][0], &mag[1][0], DFT_SIZE_W);
        lsi             = 1;
        lead_x[lsi].xcv = fCrossCorrelation (&trail_spec_nom[0], &mag[lsi][tsi_start+ioffW], trail_spec_len_nom, tsi_count, &lead_x[lsi].offi);
        ulCCCount2++;
        mi              = 0;
        while ((mi<num_modes_l) && (lss==0x01))
        {
            if (lead_x[lsi].xcv > xcthld)
                {
                    lead_x[lsi].nlm= iWFindLocalMax(&mag[lsi][0], &lead_x[lsi].lmfreq[0], &lead_x[lsi].lmqual[0], &lead_x[lsi].lmidx[0], &lead_x[lsi].lmbm);
                    if (lead_x[lsi].nlm >= nsc_min)
                    {
                        fmatch = lm_match3(&lead_x[lsi].lmfreq[0], &lead_x[lsi].lmidx[0], lead_x[lsi].nlm, &fmatchi[0], 8.0*fdthld_nom);
                        dummyf1= 0.0;
                        for (ji=0; ji< lead_x[lsi].nlm; ji++)
                            dummyf1 += lead_x[lsi].lmqual[ji];
                        r53[lsi]= (float)(fmatch)*dummyf1/(float)(lead_x[lsi].nlm);
                        if (fmatch>= nsc_min)
                        {jfm=0;
                         for (ji=0; ji<lead_x[lsi].nlm; ji++)
                         {if(lead_x[lsi].lmidx[ji] == fmatchi[jfm])
                          lead_x[lsi].lmfreq[jfm++]= lead_x[lsi].lmfreq[ji];
                         }
                         switch(lead_x[lsi].state)
                            {
                            case 0:
                                err = iCalcBestFit(&lead_x[lsi].lmfreq[0], &fmatchi[0], fmatch, &lead_x[lsi].slope, &lead_x[lsi].off_hz);
                                if (err < 0)
                                    {err_loc=4; ERN=10; goto exit_j;}
                                lead_x[lsi].state   = 1;
                                lead_x[lsi].sidx    = ulSourceIndex;
                                lead_x[lsi].r53     = r53[lsi];
                                lead_x[lsi].mode    = mi;
                                lead_x[lsi].lmcount = fmatch;
                                for (ji=0; ji<fmatch; ji++)
                                    lead_x[lsi].lmgoodi[ji] = fmatchi[ji];
                            break;

                            case 1:
                                if (r53[lsi] > lead_x[lsi].r53)
                                {
                                    err = iCalcBestFit(&lead_x[lsi].lmfreq[0], &fmatchi[0], fmatch, &lead_x[lsi].slope, &lead_x[lsi].off_hz);
                                    if (err <0)
                                        {err_loc=5; ERN=10; goto exit_j;}
                                    lead_x[lsi].sidx= ulSourceIndex;
                                    lead_x[lsi].r53= r53[lsi];
                                    lead_x[lsi].mode= mi;
                                    lead_x[lsi].lmcount=fmatch;
                                    for (ji=0; ji<fmatch; ji++)
                                        lead_x[lsi].lmgoodi[ji]= fmatchi[ji];
                                }
                                else
                                {
                                    lead_x[lsi].state=2;
                                }
                            break;

                            case 2:
                                if (r53[lsi] > lead_x[lsi].r53)
                                {
                                    err = iCalcBestFit(&lead_x[lsi].lmfreq[0], &fmatchi[0], fmatch, &lead_x[lsi].slope, &lead_x[lsi].off_hz);
                                    if (err <0)
                                        {err_loc=6; ERN=10; goto exit_j;}
                                    lead_x[lsi].sidx    = ulSourceIndex;
                                    lead_x[lsi].r53     = r53[lsi];
                                    lead_x[lsi].mode    = mi;
                                    lead_x[lsi].lmcount = fmatch;
                                    for (ji=0; ji<fmatch; ji++)
                                        lead_x[lsi].lmgoodi[ji] = fmatchi[ji];
                                }
                                else
                                {
                                    lead_x[lsi].state  = 3;
                                    lss               |= t1_FOUND ;
                                }
                            break;

                            } // end of switch lead_dat.state
                        } // end of fmatch great enough
                    } // enough local maximums
                } // end of cross-correlation value great enough
                else
                {
                    lsi=1;
                    switch(lead_x[lsi].state)
                    {
                    case 1:
                    case 2:
                        lead_x[lsi].state=3;
                        lss |= t1_FOUND ;
                    } // end of switch
                } // end of cross-correlation value below threshold
                mi++;
            } // end of while mi < num_modes_l
            ulSourceIndex-=sstepW;
        }    // end of while (looking for trailer)

        //-------------------------------------------------------------------
        if (uiDebugLevel>10)
        {
            // for now, print results to stdout

            fprintf(stdout,"\nSample index  ---------- trailer --------------------\n");
            fprintf(stdout,"              r53 val    offset (Hz)     freq. scale   mode index\n");
            fprintf(stdout,"%8ld     %7.4f     %7.2f           %9.6f      %2d\n",
                    lead_x[1].sidx, lead_x[1].r53, lead_x[1].off_hz, lead_x[1].slope, lead_x[1].mode);
        }
        //-------------------------------------------------------------------

        if (lss!= 0x03)
            {ERN=11; goto exit_j;}

        f_slope     = lead_x[0].slope;
        f_intercept = lead_x[0].off_hz;
        demod_lcnt  = 0;
        delay       = delay_nom;
        ns          = (int)(ave_nom);
        NS          = (double)(ave_nom);
        cppblk      = cppblk_nom/f_slope;    // set clock periods per block

        // found end of message
        dmd_lidx    = lead_x[1].sidx - (dmoff + lead_x[0].sidx)- (delay+ns) + dmoffW;

        // calculate expected number of framing sequences
        num_frsq    = (int)(ceil((double)(dmd_lidx/cppblk)));

        // allocate memory to store framing sequence mean squared errors
        frsqmse1    = malloc((size_t)(num_frsq*sizeof(float)));
        if(frsqmse1==NULL)
            {ERN=2; goto exit_j;}

        // allocate memory to store positions for each framing sequence
        for (j=0; j<NUM_TOTAL_SC; j++)
        {
            fridxadj[j] = malloc((size_t)(num_frsq*sizeof(int)));
            if (fridxadj[j]==NULL)
            {
                j--;
                while (j >= 0)
                    free (fridxadj[j--]);
                free (frsqmse1);
                ERN = 2;
                goto exit_j;
            }
        }

        // allocate memory to store "good/bad" indication for each framing sequence
        for (j=0; j<NUM_TOTAL_SC; j++)
        {
            fridxbetter[j] = malloc((size_t)(num_frsq*sizeof(unsigned char)));
            if (fridxbetter[j]==NULL)
            {
                j--;
                while (j >= 0)
                    free(fridxbetter[j--]);
                free(frsqmse1);
                ERN=2;
                goto exit_j;
            }
        }


DEMOD:
        // set local oscillator frequencies
        for (lo=0; lo<NUM_TOTAL_SC; lo++)
            pm_ft_ctr[lo]= pm_nom_ctr[lo]*f_slope + f_intercept;

        cppblk = cppblk_nom/f_slope;    // set clock periods per block

        //-------------------------------------------------------------------
        if (uiDebugLevel > 0)
        {
            fprintf(stdout,"\tclock periods per block is: %10.3f\n", cppblk);
        }
        //-------------------------------------------------------------------

        // demodulate each subcarrier
        for (lo=0; lo<NUM_TOTAL_SC; lo++)
        {
            FpScratchFile1 = fopen(&scratch1[0], "wb");
            if (FpScratchFile1==NULL)
            {
                fprintf(stderr,"Could not open file %s\n", &scratch1[0]);
                ERN = 8;
                goto exit_j;
            }
            FpScratchFile2 = fopen(&scratch2[0], "wb");
            if (FpScratchFile2==NULL)
            {
                fprintf(stderr,"Could not open file %s\n", &scratch2[0]);
                ERN = 8;
                fclose(FpScratchFile1);
        remove(&scratch1[0]);
                goto exit_j;
            }

            ulSourceIndex = dmoff + lead_x[0].sidx;
            end_of_file   = 0;
            ca            = 0.0;
            astep         = pm_ft_ctr[lo] * PI2 * clk_period;
            bias[lo]      = 0.0;

            while (end_of_file == 0)
            {
                val_to_read = s0blen;
                dum_len     = 0;

                if (ulSourceIndex < s0off)
                {
                    dum_len = (int)(s0off-ulSourceIndex);
                    for (si=0; si < dum_len; si++)
                        samsi[si]= dumd;
                    val_to_read -= dum_len;

                    if (fseek(FpInputFile, ulHeaderLengthB+0L, SEEK_SET)!=0)
                        {ERN=4; fclose(FpScratchFile1); fclose(FpScratchFile2); goto exit_j1;}

                    ang_ref = 0.0;
                    ang_off = astep*(double)(dum_len);
                    ca      = ang_ref-ang_off;
                    ca      = fmod(ca, PI2);
                    if (ca < 0.)
                        ca += PI2;
                }
                else
                {
                    if (fseek(FpInputFile, ulHeaderLengthB+(ulSourceIndex-s0off)*sizeof(W2B_INPUT_VAR_TYPE), SEEK_SET)!=0)
                        {ERN=4; fclose(FpScratchFile1); fclose(FpScratchFile2); goto exit_j1;}
                }
                val_to_write = fap;
                val_read     = fread(&samsi[dum_len], sizeof(W2B_INPUT_VAR_TYPE), val_to_read, FpInputFile);
                if (val_read < val_to_read)
                {
                    if (val_read < s0off)
                        {end_of_file=1; val_to_write=0;}
                    else
                        {val_to_write = (val_read > (s0off+fap) ? fap: (val_read - s0off));}
                    for (si=val_read; si< val_to_read; si++)
                        samsi[si]= dumd;
                }

                // translate to baseband, for data in samsi
                // multiply by sin and cos of local oscillator

                for (si=0; si< s0blen; si++)
                {
                    Lo1c[si]  = (float)((double)(samsi[si])*cos(ca));
                    Lo1s[si]  = (float)((double)(samsi[si])*sin(ca));
                    ca       += astep;
                    if (ca > PI2)
                        ca-= PI2;
                }

                // do decimation
                ds1t (&Lo1c[ds1dly], Lo1cf, s1blen);
                ds2t (&Lo1cf[ds2dly], Lo2cs1, s2blen);
                ds3t (&Lo2cs1[ds3dly], Lo2cs2, s3blen);
                ds1t (&Lo1s[ds1dly], Lo1sf, s1blen);
                ds2t (&Lo1sf[ds2dly], Lo2ss1, s2blen);
                ds3t (&Lo2ss1[ds3dly], Lo2ss2, s3blen);

                // do interpolation
                di=I40xxPI(&Lo2cs2[is1off], Ssbc, ipts);
                if (di < 0)
                    {ERN=32; fclose(FpScratchFile1); fclose(FpScratchFile2);  goto exit_j1;}
                di=I40xxPI(&Lo2ss2[is1off], Ssbs, ipts);
                if (di < 0)
                    {ERN=32; fclose(FpScratchFile1); fclose(FpScratchFile2); goto exit_j1;}

                // calculate angle, and remove discontinuities
                for (si=0; si< fap; si++)
                {
                    Ampd = -atan2((double)(Ssbs[si]), (double)(Ssbc[si]));
                    if (ulSourceIndex== (dmoff + lead_x[0].sidx))
                        {pv[lo]=Ampd;}
                    if ((Ampd+bias[lo] -pv[lo]) < -PI)
                        bias[lo] += PI2;
                    else
                    {
                        if ((Ampd+bias[lo]-pv[lo]) > PI)
                            bias[lo] -= PI2;
                    }
                    pv[lo]  = Ampd + bias[lo];
                    Amp[si] = (float)(pv[lo]);
                }
                val_written= fwrite(&Amp[0], sizeof(float), val_to_write, FpScratchFile1);
        if (val_written != (size_t)(val_to_write))
        {ERN=5; fclose(FpScratchFile1); fclose(FpScratchFile2); goto exit_j1;}

                // calculate Amplitude
                for (si=0; si< fap; si++)
                {
                    buf2[si]=(float) (sqrt (    (double) (Ssbs[si]*Ssbs[si] + Ssbc[si]*Ssbc[si]) ));
                }
                val_written= fwrite(&buf2[0], sizeof(float), val_to_write, FpScratchFile2);
        if (val_written != (size_t)(val_to_write))
        {ERN=5; fclose(FpScratchFile1); fclose(FpScratchFile2); goto exit_j1;}

                ulSourceIndex += fap;

                // adjust angle of local oscillators
                ang_ref = astep*(double)(ulSourceIndex);
                ang_off = astep*(double)(s0off);
                ca      = ang_ref-ang_off;
                ca      = fmod(ca, PI2);
                if (ca < 0.)
                    ca += PI2;

                // 99feb03
                //      ca2 advances fap steps per loop with no overlap or gap
                //      from loop to loop (unlike ca, which has a lot of overlap)

            } // end of while end_of_file

            fflush (FpScratchFile1);
            fclose (FpScratchFile1);
            fclose (FpScratchFile2);

            NORM = PI2/sqrt(phased2);

            // calculate and store (in a separate buffer for each lo) the average difference
            // data for the w2bscr1.tmp file just written
            if(lo==0)    // if first pass
            {
                statirr = stat (&scratch1[0], &statir);
                if (statirr!=0)
                    {ERN=6 ; goto exit_j;}
                ulFileLength_dmd       = (unsigned long) (statir.st_size);
                ulNumberOfElements_dmd = ulFileLength_dmd/(sizeof(float));

                if(demod_lcnt==0)
                {
                    // allocate buffers
                    for (j=0; j<NUM_TOTAL_SC; j++)
                    {
                        dif7n[j]=malloc((size_t)(ulNumberOfElements_dmd*sizeof(float)));
                        if (dif7n[j]==NULL)
                        {
                            j--;
                            while(j>=0)
                                free(dif7n[j--]);
                            ERN=2;
                            goto exit_j;
                        }
                    }
                    for (j=0; j<NUM_TOTAL_SC; j++)
                    {
                        dmamp[j] = malloc((size_t)(ulNumberOfElements_dmd*sizeof(float)));
                        if (dmamp[j]==NULL)
                        {
                            j--;
                            while (j>=0)
                                free(dmamp[j--]);
                            for (j=0; j<NUM_TOTAL_SC; j++)
                            free(dif7n[j]);
                            ERN=2;
                            goto exit_j;
                        }
                    }
                }

                ERN = 2;
                if ((fbufin=malloc((size_t)((ulNumberOfElements_dmd)* sizeof(float)))) == NULL)
                    goto exit_k;
                if ((fbufot=malloc((size_t)((ulNumberOfElements_dmd)* sizeof(float) ))) == NULL)
                    {free(fbufin); goto exit_k;}
            } // end of lo==0

            ERN = 7;
            FpScratchFile1 =fopen(&scratch1[0], "rb");
            if (FpScratchFile1==NULL)
            {
                fprintf(stderr,"Could not open file %s\n", scratch1);
                free(fbufin);
                free(fbufot);
                goto exit_k;
            }
            FpScratchFile2 = fopen (&scratch2[0], "rb");
            if (FpScratchFile2==NULL)
            {
                fprintf (stderr,"Could not open file %s\n", scratch2);
                fclose (FpScratchFile1);
                free (fbufin);
                free (fbufot);
                goto exit_k;
            }

            ulInputElRead = fread (&fbufin[0], sizeof(float), ulNumberOfElements_dmd, FpScratchFile1);
            if (ulInputElRead!=ulNumberOfElements_dmd)
                {ERN=13 ; fclose(FpScratchFile1); free(fbufin); free(fbufot); goto exit_k;}

            ulInputElRead = fread (&fbufot[0], sizeof(float), ulNumberOfElements_dmd, FpScratchFile2);
            if (ulInputElRead!=ulNumberOfElements_dmd)
                {ERN=13 ; fclose(FpScratchFile2); fclose(FpScratchFile1); free(fbufin); free(fbufot); goto exit_k;}

            dmd_cnt = 0;
            for (ulFIndeXB=delay+ns; ulFIndeXB<ulNumberOfElements_dmd; ulFIndeXB++)
            {
                acum1 = 0.0;
                for (si=1-ns; si<1; si++)
                    acum1 += fbufot[ulFIndeXB+si];
                dmamp[lo][dmd_cnt++]= acum1/NS ;
            }

            dmd_cnt = 0;
            for (ulFIndeXB=delay-ns+lp1dly; ulFIndeXB<ulNumberOfElements_dmd; ulFIndeXB++)
                fbufot[dmd_cnt++]= (fbufin[ulFIndeXB] - fbufin[ulFIndeXB-delay])/NORM;
            lpf01(&fbufot[lp1dly], &dif7n[lo][0], dmd_cnt-lp1len+1);

            fclose (FpScratchFile1);
            fclose (FpScratchFile2);
        } // end of for lo (demodulation)

        free (fbufin);
        free (fbufot);
        dmd_cnt = dmd_cnt - lp1len + 1;

        //-------------------------------------------------------------------
        if (uiDebugLevel > 30)
        {
            fprintf(stdout," %8d demodulated values calculated.  End of message at index %8d\n", dmd_cnt, dmd_lidx);
        }
        //-------------------------------------------------------------------

        // calculate starting point for search
        dmd_fidx = dmd_lidx -(PM_FRAME_LENGTH -1) -(int)((float)((num_frsq-1))*cppblk) -(frsq_srsam-1);
	if (dmd_fidx <0) dmd_fidx=0;

        //-------------------------------------------------------------------
        if (uiDebugLevel > 10)
        {
            fprintf(stdout,"Expecting to find %3d framing sequences\n", num_frsq);
            fprintf(stdout," Starting search at index %6d\n", dmd_fidx);
        }
        //-------------------------------------------------------------------

        // search for framing sequences
        for (lo=0; lo<NUM_TOTAL_SC; lo++)
        {
            for (si=dmd_fidx; si<dmd_fidx+frsq_srsam; si++)
            {
                acum0d = 1.0;
                sumx   = 0.0;
                for (j=0; j<num_frsq; j++)
                {
                    acum1 = 0.0;
                    di    = si + (int)((float)(j)*cppblk);
                    for (ki=0; ki<PM_FRAME_LENGTH; ki++)
                    {
                        dummyf1  = dif7n[lo][di+ki]- pm_frame[ki];
                        acum1   += dummyf1*dummyf1;
                    } // end of for ki
                    frsqmse1[j]  = acum1/(float)(PM_FRAME_LENGTH);
                    sumx        += frsqmse1[j];
                    if(j==0)
                        {sumy= frsqmse1[j]; frsqmse1i=j;}
                    else
                        if (frsqmse1[j] > sumy)
                            {sumy= frsqmse1[j]; frsqmse1i=j;}
                } // end of for j

                // calculate product of mean and standard deviation, leaving out largest value
                sumx  = sumx - frsqmse1[frsqmse1i];
                sumx  = sumx/(float)(num_frsq -1);
                sumx2 = 0.0;
                for (j=0; j<num_frsq; j++)
                {
                    if (j==frsqmse1i)
                        continue;
                    sumx2 = sumx2 + (frsqmse1[j] - sumx) * (frsqmse1[j] - sumx);
                } // end of for j

                acum0d = (double) (sumx) * sqrt((double)(sumx2/(double)(num_frsq-2)));
                if (si == dmd_fidx)
                {
                    frsq_mmse[lo]  = acum0d;
                    fridxoff[lo]   = si-dmd_fidx;
                    frsqmsesd[lo]  = (float)(acum0d)/sumx;
                    frsqmseave[lo] = sumx;
                }
                else
                {
                    if (acum0d < frsq_mmse[lo])
                    {
                        frsq_mmse[lo]  = acum0d;
                        fridxoff[lo]   = si-dmd_fidx;
                        frsqmsesd[lo]  = (float)(acum0d)/sumx;
                        frsqmseave[lo] = sumx;
                    }
                }
            } // end of for si
        } // end of for lo searching for framing sequences

        //-------------------------------------------------------------------
        if (uiDebugLevel > 10)
        {
            fprintf(stdout,"Results of initial search for framing sequences\n");
            fprintf(stdout,"  sub-carrier  offset   min product\n");
            for (lo=0; lo<NUM_TOTAL_SC; lo++)
                fprintf(stdout,"     %1d      %6d      %10.5g\n", lo, fridxoff[lo], frsq_mmse[lo]);
        }
        //-------------------------------------------------------------------

        // determine clock periods per block, from best framing sequence data
        // don't use data from subcarrier 0, due to non-linear phase filters in radios

        // set index of best framing sequence

        // use best (except for lowest freq) overall subcarrier for next step
        bfridx     = 1;
        bfrsq_mmse = frsq_mmse[1];
        for (lo=2; lo<NUM_TOTAL_SC; lo++)
        {
            if (frsq_mmse[lo]< bfrsq_mmse)
                {bfrsq_mmse= frsq_mmse[lo]; bfridx=lo;}
        }

        // use best framing sequence for this subcarrier as the reference
        for (j=0; j<num_frsq; j++)
        {
            acum1 = 0.0;
            di    = fridxoff[bfridx]+dmd_fidx + (int)((float)(j)*cppblk);
            for (ki=0; ki<PM_FRAME_LENGTH; ki++)
            {
                dummyf1  = dif7n[bfridx][di+ki]- pm_frame[ki];
                acum1   += dummyf1*dummyf1;
            } // end of for ki
            acum1 = acum1/(float)(PM_FRAME_LENGTH);
            if(j==0)
                {sumy= acum1; bfridx0=j;}
            else
                if (acum1 < sumy)
                    {sumy= acum1; bfridx0=j;}
        } // end of for j

        fridxadj[bfridx][0] = fridxoff[bfridx] + dmd_fidx;
        for (j=1; j<num_frsq; j++)
            fridxadj[bfridx][j] = fridxadj[bfridx][0]+ (int)(floor((double)((float)(j)*cppblk + 0.5)));

        // search for a better value for cppblk
        lo = bfridx;
        for (si=-cppblk_ossr; si<=cppblk_ossr; si++)
        {
            acum0d = 1.0;
            sumx   = 0.0;
            for (j=0; j<num_frsq; j++)
            {
                acum0 = (float)(j-bfridx0)*(cppblk+(float)(si));
                bk    = fridxadj[bfridx][bfridx0] + (int)(floor((double)(acum0+0.5)));
                acum1 = 0.0;
                for (ki=0; ki<PM_FRAME_LENGTH; ki++)
                {
                    dummyf1  = dif7n[lo][bk+ki]- pm_frame[ki];
                    acum1   += dummyf1*dummyf1;
                } // end of for ki
                frsqmse1[j]  = acum1/(float)(PM_FRAME_LENGTH);
                sumx        += frsqmse1[j];
                if(j==0)
                    {sumy= frsqmse1[j]; frsqmse1i=j;}
                else
                    if (frsqmse1[j] > sumy)
                        {sumy= frsqmse1[j]; frsqmse1i=j;}
            } // end of for j

            // calculate product of mean and standard deviation, leaving out largest value
            sumx  = sumx - frsqmse1[frsqmse1i];
            sumx  = sumx/(float)(num_frsq -1);
            sumx2 = 0.0;
            for (j=0; j<num_frsq; j++)
            {
                if(j==frsqmse1i)
                    continue;
                sumx2 = sumx2 + (frsqmse1[j] - sumx) * (frsqmse1[j] - sumx);
            } // end of for j
            acum0d = (double)(sumx)* sqrt(     (double)(sumx2/(double)(num_frsq-2)) );
            if (si== -cppblk_ossr)
                {frsq_mmse[lo]= acum0d; fridxoffsr[lo]= si;}
            else
            {
                if (acum0d < frsq_mmse[lo])
                    {frsq_mmse[lo]= acum0d; fridxoffsr[lo]= si;}
            }
        } // end of for si

        if (fridxoffsr[lo]==0)
            cppblk_fr = cppblk;
        else
            cppblk_fr = cppblk+(float)(fridxoffsr[lo]);

        //-------------------------------------------------------------------
        if (uiDebugLevel > 20)
        {
            fprintf(stdout,"Best min product found is: %10.5g, using an offset of %3d\n", frsq_mmse[lo], fridxoffsr[lo]);
            fprintf(stdout,"        Now using %10.3f clock periods per block\n", cppblk_fr);
        }
        //-------------------------------------------------------------------

        // refine framing sequence positions, based on actual data, for best subcarrier
        for (j=0; j<num_frsq; j++)
        {
            for (ki= -frosftr2; ki<= frosftr2; ki++)
            {
                acum1 = 0.0;
                for (di=0; di<PM_FRAME_LENGTH; di++)
                {
                    dummyf1  = dif7n[bfridx][fridxadj[bfridx][j]+di+ki]- pm_frame[di];
                    acum1   += dummyf1*dummyf1;
                } // end of for di
                if (ki== -frosftr2)
                    {acum0= acum1; bk =ki;}
                else
                {
                    if (acum1 < acum0)
                        {acum0= acum1; bk =ki;}
                }
                if (ki== 0)
                    msenom=acum1;
            } // end of for ki

            if ((acum0/(float)(PM_FRAME_LENGTH)) < frsqmseave[bfridx])
                {fridxadj[bfridx][j] += bk ; fridxbetter[bfridx][j]= 0x01;}
            else
                {fridxbetter[bfridx][j]= 0x00;}
        } // end of for j

        // calculate least squares best fit straight line through the "better than average" framing sequence positions
        sumx  = 0.0;
        sumy  = 0.0;
        sumxy = 0.0;
        x2sum = 0.0;
        ki    = 0;
        for (j=0; j<num_frsq; j++)
        {
            if (fridxbetter[bfridx][j] == 0x01)
            {
                cx     = (float)(j);
                sumx  += cx; x2sum += cx*cx;
                sumxy += cx*(float)(fridxadj[bfridx][j]);
                sumy  += (float)(fridxadj[bfridx][j]);
                ki++;
            }
        }
        if (ki > 2)
        {
            cppblk_slope[bfridx] = ((float)(ki)*sumxy - sumx*sumy)/((float)(ki)*x2sum - sumx*sumx);
            cppblk_off[bfridx]   = sumy/(float)(ki) - cppblk_slope[bfridx]*sumx/(float)(ki);
        }
        else
        {
            cppblk_slope[bfridx] = cppblk_fr;
            cppblk_off[bfridx]   = dmd_fidx + fridxoff[bfridx];
        }

        // if offset is different from 0, and this is the first pass, demodulate again
        if (fridxoffsr[lo] != 0)
        {
            if (demod_lcnt<1)
                {
                    f_slope = (float)(cppblk_nom)/cppblk_slope[bfridx];

                    //-------------------------------------------------------
                    if (uiDebugLevel > 20)
                    {
                        fprintf(stdout,"Starting pass number %1d through the demodulator.\n", demod_lcnt+2);
                    }
                    //-------------------------------------------------------

                    demod_lcnt++;
                    goto DEMOD;
                }
            }

            // calculate expected position for all framing sequences for all subcarriers,
            // based on best framing sequence for best subcarrier
            for (j=0; j<num_frsq; j++)
            {
                acum0 = (float)(j-bfridx0)*cppblk_fr;
                bk = fridxadj[bfridx][bfridx0] - soff[bfridx] + (int)(floor((double)(acum0+0.5)));
                for (lo=0; lo<NUM_TOTAL_SC; lo++)
                    fridxadj[lo][j] = bk+ soff[lo];
            }

            // take care of lowest frequency subcarrier as a special case
            lo = 0;
            for (si=-36; si< 48; si++)
            {
                acum0d = 1.0;
                sumx   = 0.0;
                for (j=0; j<num_frsq; j++)
                {
                    acum1 = 0.0;
                    di    = si+ fridxadj[lo][j];
                    for (ki=0; ki<PM_FRAME_LENGTH; ki++)
                    {
                        dummyf1  = dif7n[lo][di+ki]- pm_frame[ki];
                        acum1   += dummyf1*dummyf1;
                    } // end of for ki
                    frsqmse1[j]  = acum1/(float)(PM_FRAME_LENGTH);
                    sumx        += frsqmse1[j];
                    if (j==0)
                        {sumy= frsqmse1[j]; frsqmse1i=j;}
                    else
                        if (frsqmse1[j] > sumy)
                            {sumy= frsqmse1[j]; frsqmse1i=j;}
                } // end of for j

                // calculate product of mean and standard deviation, leaving out largest value
                sumx  = sumx - frsqmse1[frsqmse1i];
                sumx  = sumx/(float)(num_frsq -1);
                sumx2 = 0.0;
                for (j=0; j<num_frsq; j++)
                {
                    if (j==frsqmse1i)
                        continue;
                    sumx2 = sumx2 + (frsqmse1[j] - sumx) * (frsqmse1[j] - sumx);
                } // end of for j

                acum0d = (double)(sumx)* sqrt(    (double)(sumx2/(double)(num_frsq-2)));
                if (si== -36)
                {
                    frsq_mmse[lo]  = acum0d;
                    fridxoffsr[lo] = si;
                }
                else
                {
                    if (acum0d < frsq_mmse[lo])
                        {frsq_mmse[lo]= acum0d; fridxoffsr[lo]= si;}
                }
            } // end of for si

            if (fridxoffsr[lo] != 0)
                {for (j=0; j<num_frsq; j++) fridxadj[lo][j] += fridxoffsr[lo];}

            // refine framing sequence positions, based on actual data
            for (j=0; j<num_frsq; j++)
            {
                for (lo=0; lo<NUM_TOTAL_SC; lo++)
                {
                    for (ki= -frosftr2; ki<= frosftr2; ki++)
                    {
                        acum1 = 0.0;
                        for (di=0; di<PM_FRAME_LENGTH; di++)
                        {
                            dummyf1  = dif7n[lo][fridxadj[lo][j]+di+ki]- pm_frame[di];
                            acum1   += dummyf1*dummyf1;
                        } // end of for di
                        if (ki== -frosftr2)
                            {acum0= acum1; bk =ki;}
                        else
                        {
                            if (acum1 < acum0)
                                {acum0= acum1; bk =ki;}
                        }
                        if (ki== 0)
                            msenom=acum1;
                    } // end of for ki

                    if ((acum0/(float)(PM_FRAME_LENGTH)) < mseavexf*frsqmseave[lo])
                        {fridxadj[lo][j] += bk ; fridxbetter[lo][j]= 0x01;}
                    else
                        {fridxbetter[lo][j]= 0x00;}
                } // end of for lo
            } // end of for j

            free(frsqmse1);

            //---------------------------------------------------------------
            if  (uiDebugLevel > 20)
            {
                fprintf(stdout,"Better than average cases => 1,  average, or worse => 0\n");
                fprintf(stdout," framing --------- sub-carrier index----------\n");
                fprintf(stdout,"  index   0    1    2    3    4    5    6    7\n");
                for (j=0; j<num_frsq; j++)
                {
                    fprintf(stdout,"   %3d  ", j);
                    for (lo=0; lo<NUM_TOTAL_SC; lo++)
                        fprintf(stdout,"  %1c  ", fridxbetter[lo][j]+ 0x30);
                    fprintf(stdout,"\n");
                }
            }
            //---------------------------------------------------------------

            // calculate least squares best fit straight line through the "better than
            // average" framing sequence positions
            scv = 0;    // to count number of "valid subcarriers"
            for (lo=0; lo<NUM_TOTAL_SC; lo++)
            {
                sumx  = 0.0;
                sumy  = 0.0;
                sumxy = 0.0;
                x2sum = 0.0;
                ki    = 0;
                for (j=0; j<num_frsq; j++)
                {
                    if (fridxbetter[lo][j] == 0x01)
                    {
                        cx     = (float)(j);
                        sumx  += cx; x2sum += cx*cx;
                        sumxy += cx*(float)(fridxadj[lo][j]);
                        sumy  += (float)(fridxadj[lo][j]);
                        ki++;
                    }
                }
                if (ki > 2)
                {
                    slope_valid[scv++] = lo;
                    cppblk_slope[lo]   = ((float)(ki)*sumxy - sumx*sumy)/((float)(ki)*x2sum - sumx*sumx);
                    cppblk_off[lo]     = sumy/(float)(ki) - cppblk_slope[lo]*sumx/(float)(ki);
                }
                else
                {
                    cppblk_slope[lo] = 0.0;
                    cppblk_off[lo]   = dmd_fidx + fridxoff[lo];
                }
            } // end of for lo


            //---------------------------------------------------------------
            if (uiDebugLevel > 20)
            {
                fprintf(stdout,"Results of best fit straight lines through better points\n");
                fprintf(stdout,"  sub-carrier    offset         slope\n");
                for (lo=0; lo<NUM_TOTAL_SC; lo++)
                    fprintf(stdout,"     %1d   %12.7g      %12.7g\n", lo, cppblk_off[lo], cppblk_slope[lo]);
            }
            //---------------------------------------------------------------

            // leave out highest and lowest, use average of remaining slopes for subsequent calculations
            if (scv > 2)
            {
                sumy = (sumx = cppblk_slope[slope_valid[0]]);

                for (j=1; j<scv; j++)                        // find sum and max
                {
                    sumx += cppblk_slope[slope_valid[j]];
                    if (cppblk_slope[slope_valid[j]] > sumy)
                        sumy = cppblk_slope[slope_valid[j]];
                }
                sumx = sumx - sumy;
                sumy = cppblk_slope[slope_valid[0]];        // find min
                for (j=1; j<scv; j++)
                {
                    if (cppblk_slope[slope_valid[j]] < sumy)
                        sumy = cppblk_slope[slope_valid[j]];
                }
                sumx      = sumx - sumy;
                cppblk_ft = sumx/(float)(scv -2);
            }
            else
            {
                sumx = cppblk;
                if (scv > 0)
                {
                    for (j=0; j< scv; j++)
                        sumx += cppblk_slope[slope_valid[j]];
                    cppblk_ft = sumx/(float)(scv+1);
                }
                else
                    cppblk_ft = cppblk;
            }

            if (cppblk_ft > cppblk_max)            // limit cppblk_ft
                cppblk_ft = cppblk_max;
            else
            {
                if (cppblk_ft < cppblk_min)
                    cppblk_ft = cppblk_min;
            }

            //---------------------------------------------------------------
            if (uiDebugLevel > 0)
            {
                fprintf(stdout,"        Now using %12.7g clock periods per block\n", cppblk_ft);
            }
            //---------------------------------------------------------------

            // limit cppblk_off values
            for (lo=0; lo<NUM_TOTAL_SC; lo++)
            {
                if (lo==bfridx)
                    continue;
                if (((cppblk_off[lo] - dmd_fidx) > (fridxoff[lo] + cppblk_off_lim)) ||
                 ((cppblk_off[lo] - dmd_fidx) < (fridxoff[lo] - cppblk_off_lim)))
                {
                    cppblk_off[lo]= cppblk_off[bfridx] - soff[bfridx] + soff[lo];
                }
            }

            // substitute for the average or worse cases, framing sequence positions
            // calculated from the best fit straight lines
            for (lo=0; lo<NUM_TOTAL_SC; lo++)
            {
                for (j=0; j<num_frsq; j++)
                {
                    if (fridxbetter[lo][j] == 0x00)
                        fridxadj[lo][j] = (int) (floor((double)(cppblk_off[lo] + cppblk_ft*(float)(j)) + 0.5));
                }
            } // end of for lo

            //---------------------------------------------------------------
            if (uiDebugLevel > 30)
            {
                fprintf(stdout," framing   --------------------- sub-carrier index------------------\n");
                fprintf(stdout,"  index    0       1       2       3       4       5       6       7\n");
                for (j=0; j<num_frsq; j++)
                {
                    fprintf(stdout,"   %3d  ", j);
                    for (lo=0; lo<NUM_TOTAL_SC; lo++)
                        fprintf(stdout,"%6d  ", fridxadj[lo][j]);
                    fprintf(stdout,"\n");
                }
            }
            //---------------------------------------------------------------

            //---------------------------------------------------------------
            if (uiDebugLevel > 40)
            {
                // for testing purposes dump demodulated signals
                for (lo=0; lo<NUM_TOTAL_SC; lo++)
                {
                    FpOutputFile = NULL;
                    FpOutputFile = fopen (lodif7nfn[lo], "wt");
                    if (FpOutputFile==NULL)
                    {
                        fprintf(stderr,"Could not open file %s\n", lodif7nfn[lo]);
                        goto exit_k;
                    }

                    if (uiDebugLevel < 50)
                    {
                        for (si=fridxadj[2][0]-10*cppsym; si<dmd_lidx; si++)
                            fprintf(FpOutputFile,"%8.4f\n", dif7n[lo][si]);
                    }
                    else
                    {
                        for (si=fridxadj[2][0]-10*cppsym; si<dmd_lidx; si++)
                            fprintf(FpOutputFile,"%8.4f\n", dmamp[lo][si]);
                    }

                    fclose(FpOutputFile);
                }
            }
            //---------------------------------------------------------------

            // determine error correction mode (# of redundancy symbols per block)
            for (si=0; si<num_ec_lvls; si++)
                ec_mode_dist[si]=0.0;

            // check three pair of symbols preceding 1st framing sequence
            for (si= -6*cppsym; si<0; si+= 2*cppsym)
            {
                for (lo=0; lo<NUM_TOTAL_SC; lo++)
                {
                    c2phbuf[lo] = dif7n[lo][si+fridxadj[lo][0]];
                }
                di = decode2ed1mode (&c2phbuf[0], &distg, mode1dthld);
                if (di >= 0)
                    ec_mode_dist[di] += 1.0/(distg > 1e-3 ? distg : 1e-3);
                for (lo=0; lo<NUM_TOTAL_SC; lo++)
                {
                    c2phbuf[lo]= dif7n[lo][si+cppsym+fridxadj[lo][0]];
                }
                di = decode2ed2mode(&c2phbuf[0], &distg, mode2dthld);
                if (di >= 0)
                    ec_mode_dist[di] += 1.0/(distg > 1e-3 ? distg : 1e-3);
            } // end of for si, checking initial mode id symbols

            // check three pair of symbols following last framing sequence
            for (si= frame_len*cppsym; si<=(frame_len+4)*cppsym; si+= 2*cppsym)
            {
                for (lo=0; lo<NUM_TOTAL_SC; lo++)
                {
                    c2phbuf[lo]= dif7n[lo][si+fridxadj[lo][num_frsq-1]];
                }
                di = decode2ed1mode(&c2phbuf[0], &distg, mode1dthld);
                if (di >= 0)
                    ec_mode_dist[di] += 1.0/(distg > 1e-3 ? distg : 1e-3);
                for (lo = 0; lo < NUM_TOTAL_SC; lo++)
                {
                    c2phbuf[lo] = dif7n[lo][si+cppsym+fridxadj[lo][num_frsq-1]];
                }
                di = decode2ed2mode (&c2phbuf[0], &distg, mode2dthld);
                if (di >= 0)
                    ec_mode_dist[di] += 1.0/(distg > 1e-3 ? distg : 1e-3);
            } // end of for si, checking final mode id symbols


            //---------------------------------------------------------------
            if (uiDebugLevel > 0)
            {
                fprintf (stdout,"\nError Correction Mode Determination Weights\n");
                fprintf (stdout,"  Mode Index    Weights\n");
                for (si=0; si<num_ec_lvls; si++)
                    fprintf (stdout,"     %2d       %8.4f\n", si, ec_mode_dist[si]);
            }
            fprintf (stdout, "\n");
            //---------------------------------------------------------------

            dummyf1 = ec_mode_dist[0];
            ec_mode = 0;
            for (ki=1; ki<num_ec_lvls; ki++)
            {
                if (ec_mode_dist[ki] > dummyf1)
                    {dummyf1= ec_mode_dist[ki]; ec_mode= ki;}
            }
            if (ec_mode_dist[ec_mode]==0.0)
                ec_mode=default_mode;
            iidx_mn1 = iidx_mnx[ec_mode];
            k1       = kx[ec_mode];
            t1       = k1/2;
            bad_blks = 0;
            ERN      = 0;

            //  for each block
            //  decode inner code from demodulated phase changes to channel symbol numbers
            //  decode array of channel symbol numbers
            //  output erasure and error correction information
            //  if 1st block
            //      extract filename, and file length, open file, store remaining data in file
            //  else
            //      store all "non-dummy" data in file

            //---------------------------------------------------------------
            if (uiDebugLevel > 0)
            {
                fprintf(stdout,"Block inner  |      outer code      | outer   |       inner code\n");
                fprintf(stdout,"  #   code   |       erasures       | code    |         changes\n");
                fprintf(stdout,"      changes|1st   2nd   3rd   4th | changes |    0   1   2   3   4\n");
                fprintf(stdout,"----  ------ |---   ---   ---   --- |   ---   |  --- --- --- --- ---\n");
            }
            //---------------------------------------------------------------

            for (j=0; j< (num_frsq-1); j++)   /*??*/
            {

            //---------------------------------------------------------------
            //  refine symbol positions, based on amplitude variation information:
            //
            //  1. Fill sym_pos[][] with nominal symbol positions, based on surrounding
            //     framing sequence positions.  Fill ma1[][] with 0x00, to indicate
            //     nominal calculated positions.
            //  2. For all symbol locations with 0x00 in ma1[][], if local maximum of
            //     cross-correlation between amplitude reference pattern and actual
            //     amplitude values is greater than thld_amp1, and is within ossi_amp1
            //     index values of the nominal position, then use the location of this
            //     local maximum as the position of that symbol, and set the
            //     corresponding location in ma1[][] to 0x01.
            //  3. Extend, form both ends, each run of 0x01 values in ma1[][].  A run
            //     is extended by adjacent local maximum cross-correlation values being
            //     > thld_amp2 and within ossi_amp2 locations of the, f_slope adjusted,
            //     position from the previous symbol's position.  Set the corresponding
            //     locations in ma1[][] to 0x02, for those symbols whose position was
            //     changed in this step.
            //  4. Replace symbol positions corresponding to remaining runs of 0x00 in
            //     ma1[][] with linearly interpolated values, based on the positions
            //     corresponding to the non-0x00 entries in ma1[][] adjacent to the
            //     runs of 0x00.  Leave any end symbol positions with 0x00 values in
            //     ma1[][] as they are.  Set locations in ma1[][] corresponding to
            //     symbol positions changed in this step to 0x03.
            //---------------------------------------------------------------

                //-----------------------------------------------------------
                //  1. Fill sym_pos[][] with nominal symbol positions, based on surrounding
                //     framing sequence positions.  Fill ma1[][] with 0x00, to indicate
                //     nominal calculated positions.

                for (lo=0; lo<NUM_TOTAL_SC; lo++)
                {
                    sym_pos[lo][0] = (double)(fridxadj[lo][j]);
                    dcidx_step[lo] = ((double)(fridxadj[lo][j+1]) - sym_pos[lo][0])/ (double)(MAX_GF_EL_1 + frame_len);
                    for (si=1; si< MAX_GF_EL_1+frame_len; si++)
                        {sym_pos[lo][si]= sym_pos[lo][si-1] + dcidx_step[lo];}
                    for (si=0; si< MAX_GF_EL_1+frame_len; si++)
                        ma1[lo][si]=0x00;
                }

                //-----------------------------------------------------------
                if (uiDebugLevel > 35)
                {
                    fprintf(stderr,"#BLOCK: %2d  Step 1\n", j);
                    for (si=0; si<MAX_GF_EL_1+frame_len; si++)
                    {
                       for (lo=0; lo<NUM_TOTAL_SC; lo++)
                            fprintf(stderr,"%7.0f ", sym_pos[lo][si]);
                       for (lo=0; lo<NUM_TOTAL_SC; lo++)
                            fprintf(stderr,"%1d ", (int)(ma1[lo][si]));
                       fprintf(stderr,"\n");
                    }
                }
                //-----------------------------------------------------------

                //-----------------------------------------------------------
                //  2. For all symbol locations with 0x00 in ma1[][], if local maximum of
                //     cross-correlation between amplitude reference pattern and actual
                //     amplitude values is greater thanthld_amp1, and is within ossi_amp1
                //     index values of the nominal position, then use the location of this
                //     local maximum as the position of that symbol, and set the
                //     corresponding location in ma1[][] to 0x01.

                for (lo=0; lo<NUM_TOTAL_SC; lo++)
                {
                    for (si=0; si< MAX_GF_EL_1+frame_len; si++)
                    {
                        spni = (int) (floor(sym_pos[lo][si]+0.5));
                        sas  = iFindLocalMaxAmp(&dmamp[lo][spni], ossi_amp1, &lm_off_amp, &lm_value_amp);
                        if ((sas==1) && (lm_value_amp > thld_amp1))
                        {
                            sym_pos[lo][si] += lm_off_amp;
                            ma1[lo][si]      = 0x01;
                        }
                    } // end of for si
                } // end of for lo

                //-----------------------------------------------------------
                if (uiDebugLevel > 35)
                {
                    fprintf(stderr,"#BLOCK: %2d  Step 2\n", j);
                    for (si=0; si<MAX_GF_EL_1+frame_len; si++)
                    {
                        for (lo=0; lo<NUM_TOTAL_SC; lo++)
                            fprintf(stderr,"%7.0f ", sym_pos[lo][si]);
                        for (lo=0; lo<NUM_TOTAL_SC; lo++)
                            fprintf(stderr,"%1d ", (int)(ma1[lo][si]));
                        fprintf(stderr,"\n");
                    }
                }
                //-----------------------------------------------------------

                //-----------------------------------------------------------
                //  3. Extend, from both ends, each run of 0x01 values in ma1[][].  A run
                //     is extended by adjacent local maximum cross-correlation values being
                //     > thld_amp2 and within ossi_amp2 locations of the, f_slope adjusted,
                //     position from the previous symbol's position.  Set the corresponding
                //     locations in ma1[][] to 0x02, for those symbols whose position was
                //     changed in this step.

                for (lo=0; lo<NUM_TOTAL_SC; lo++)
                {
                    ar_state = 0;
                    aridx    = 0;
                    for (si=0; si< (MAX_GF_EL_1+frame_len); si++)  /*??*/
                    {
                    switch(ar_state)
                        {
                        case 0:
                            if (ma1[lo][aridx]==0x00)
                                {arzdx = aridx; ar_state=1;}
                            else
                                {arldx= aridx; ar_state=2;}
                        break;

                        case 1:
                            if (ma1[lo][++aridx]==0x01)
                                ar_state=3;
                        break;

                        case 2:
                            if (ma1[lo][++aridx]==0x00)
                                ar_state=4;
                        break;

                        case 3:
                            for (arudx=aridx-1; arudx >= arzdx; arudx--)
                            {
                                spni = (int) (floor(sym_pos[lo][arudx]+0.5));
                                sas  = iFindLocalMaxAmp (&dmamp[lo][spni], ossi_amp2, &lm_off_amp, &lm_value_amp);
                                if ((sas==1) && (lm_value_amp > thld_amp2))
                                {
                                    sym_pos[lo][arudx] += lm_off_amp;
                                    ma1[lo][arudx]      = 0x02;
                                }
                                else
                                    break;
                            } // end of for arudx
                            arldx    = aridx;
                            ar_state = 2;
                        break;

                        case 4:
                            spni = (int)(floor(sym_pos[lo][aridx]+0.5));
                            sas  = iFindLocalMaxAmp (&dmamp[lo][spni], ossi_amp2, &lm_off_amp, &lm_value_amp);
                            if ((sas==1) && (lm_value_amp > thld_amp2))
                            {
                                sym_pos[lo][aridx] += lm_off_amp;
                                ma1[lo][aridx]      = 0x02;
                                if (aridx < (MAX_GF_EL_1 + frame_len - 1))
                                    ar_state=2;
                                else
                                    ar_state=6;
                            }
                            else
                            {
                                if (aridx < (MAX_GF_EL_1+frame_len -1))
                                {
                                    arzdx    = aridx;
                                    ar_state = 5;
                                }
                                else
                                    ar_state = 6;
                            }
                        break;

                        case 5:
                            if (ma1[lo][++aridx]==0x00)
                            {
                                if (aridx < (MAX_GF_EL_1+frame_len-1))
                                    ar_state =1;
                                else
                                    ar_state=6;
                            }
                            else
                                ar_state=3;
                        break;

                        case 6:
                        break;

                        default:
                            error_msg_to_stderr(ERN=14);
                        } // end of switch(ar_state)
                    } // end of for si
                } // end of for lo

                //-----------------------------------------------------------
                if (uiDebugLevel > 35)
                {
                    fprintf(stderr,"#BLOCK: %2d  Step 3\n", j);
                    for (si=0; si<(MAX_GF_EL_1+frame_len); si++)
                    {
                        for (lo=0; lo<NUM_TOTAL_SC; lo++)
                            fprintf(stderr,"%7.0f ", sym_pos[lo][si]);
                        for (lo=0; lo<NUM_TOTAL_SC; lo++)
                            fprintf(stderr,"%1d ", (int)(ma1[lo][si]));
                        fprintf(stderr,"\n");
                    }
                }
                //-----------------------------------------------------------

                //-----------------------------------------------------------
                //  4. Replace symbol positions corresponding to remaining runs of 0x00 in
                //     ma1[][] with linearly interpolated values, based on the positions
                //     corresponding to the non-0x00 entries in ma1[][] adjacent to the
                //     runs of 0x00.  Leave any end symbol positions with 0x00 values in
                //     ma1[][] as they are.  Set locations in ma1[][] corresponding to
                //     symbol positions changed in this step to 0x03.

                for (lo=0; lo < NUM_TOTAL_SC; lo++)
                {
                    ar_state = 0;
                    aridx    = 0;
                    nzidxl   = 0;
                    nzidxr   = 0;
                    for (si=0; si<(MAX_GF_EL_1+frame_len); si++)
                    {
                        switch(ar_state)
                        {
                        case 0:
                            if (ma1[lo][aridx]==0x00) ar_state=2; else ar_state=1;
                        break;

                        case 1:
                            if (ma1[lo][++aridx]==0x00)
                                {nzidxl= aridx-1; ar_state=2;}
                            else
                                {if (si==(MAX_GF_EL_1+frame_len-1)) ar_state=3;}
                        break;

                        case 2:
                            if ((ma1[lo][++aridx]!=0x00) || (si==(MAX_GF_EL_1+frame_len-1)))
                            {
                                nzidxr   = aridx;
                                dummyd1  = sym_pos[lo][nzidxr] - sym_pos[lo][nzidxl];
                                ar_step0 = dummyd1/(double)(nzidxr-nzidxl);
                                for (ii0=(nzidxl+1); ii0< nzidxr; ii0++)
                                {
                                    sym_pos[lo][ii0] = sym_pos[lo][ii0-1] + ar_step0;
                                    ma1[lo][ii0]     = 0x03;
                                }
                                if (si==(MAX_GF_EL_1+frame_len-1))
                                    ar_state=3;
                                else
                                    ar_state=1;
                            }
                        break;

                        case 3:
                        break;

                        default:
                            error_msg_to_stderr(ERN=15);
                        } // end of switch(ar_state)
                    } // end of for si
                } // end of for lo

                //-----------------------------------------------------------
                if (uiDebugLevel > 35)
                {
                    fprintf(stderr,"#BLOCK: %2d  Step 4\n", j);
                    for (si=0; si<MAX_GF_EL_1+frame_len; si++)
                    {
                        for (lo=0; lo<NUM_TOTAL_SC; lo++)
                            fprintf(stderr,"%7.0f ", sym_pos[lo][si]);
                        for (lo=0; lo<NUM_TOTAL_SC; lo++)
                            fprintf(stderr,"%1d ", (int)(ma1[lo][si]));
                        fprintf(stderr,"\n");
                    }
                }
                //-----------------------------------------------------------

                eracnt1     = 0;
                scca        = 0;
                eracnt1s    = 0;
                eracnt1ss   = 0;
                eracnt1sss  = 0;
                eracnt1ssss = 0;
                for (si=0; si<NUM_TOTAL_SC; si++)
                    sccx[si]=0;
                for (si= 0; si< MAX_GF_EL_1; si++)
                {
                    for (lo=0; lo<NUM_TOTAL_SC; lo++)
                    {
                        c2phbuf[lo] = dif7n[lo][(int)(floor(sym_pos[lo][si+frame_len]+0.5))];
                    } // end of for lo

                    di = decode2wne2 (&c2phbuf[0], &scc, &ob_cnt, &ob_mask);

                    //-------------------------------------------------------
                    if (uiDebugLevel > 50)
                    {
                        fprintf(stdout,"%3d: %3d_%1d  ", si, di, scc);
                        for (lo=0; lo<NUM_TOTAL_SC; lo++)
                            fprintf(stdout,"%5.2f  ", c2phbuf[lo]);
                    }
                    //-------------------------------------------------------

                    if ((di<0) || (scc > 0))
                    {
                        if (ob_cnt > NUM_REDUNDANT_SC)    // too many out of bounds subcarrier values
                        {
                            erasi1[0][eracnt1++] = MAX_INFO_INDEX-si;    // hard erasure
                            if (di >= 0)
                                cb[0][MAX_INFO_INDEX-si] =(unsigned int)(di);
                        }
                        else
                        {
                            if (di < 0)
                            {
                            switch (ob_cnt)
                            {
                            case 0:
                            case 1:
                                di_last = decode2ed2w1e(&c2phbuf[0], &distg, (si<(MAX_GF_EL_1-iidx_mn1)?260:MAX_GF_EL_1), &erag, &scc_last, ob_mask, &bec_mask);
                            break;

                            case 2:
                                di_last = decode2ed2w2e(&c2phbuf[0], &distg, (si<(MAX_GF_EL_1-iidx_mn1)?260:MAX_GF_EL_1), &erag, &scc_last, ob_mask, &bec_mask);
                            break;

                            case 3:
                                di_last = decode2ed2w3e(&c2phbuf[0], &distg, (si<(MAX_GF_EL_1-iidx_mn1)?260:MAX_GF_EL_1), &erag, &scc_last, ob_mask, &bec_mask);
                            break;

                            case 4:
                                di_last = decode2ed2w4e(&c2phbuf[0], &distg, (si<(MAX_GF_EL_1-iidx_mn1)?260:MAX_GF_EL_1), &erag, &scc_last, ob_mask, &bec_mask);
                                if (erag==1)
                                    {erasi1[0][eracnt1++]= MAX_INFO_INDEX-si;}      // hard erasure
                                else
                                    {erasi1s[0][eracnt1s++]= MAX_INFO_INDEX-si;}    // soft erasure
                                if (di_last >= 0)
                                    cb[0][MAX_INFO_INDEX-si]=(unsigned int)(di_last);
                            break;
                        } // end of switch ob_cnt
                    } // end of di < 0
                    else
                    {
                        switch (scc)    // # of valid cases depends on amount of redundancy in inner code
                        {
                        case 1:
                            di_last = decode2ed2w2e(&c2phbuf[0], &distg, (si<(MAX_GF_EL_1-iidx_mn1)?260:MAX_GF_EL_1), &erag, &scc_last, ob_mask, &bec_mask);
                        break;

                        case 2:
                            di_last = decode2ed2w3e(&c2phbuf[0], &distg, (si<(MAX_GF_EL_1-iidx_mn1)?260:MAX_GF_EL_1), &erag, &scc_last, ob_mask, &bec_mask);
                        break;
                        } // end of switch scc
                    } // end of di >= 0

                    while ((scc_last < NUM_REDUNDANT_SC) && (di != di_last))
                    {
                        di = di_last; // save previous value
                        switch (scc_last)
                        {
                        case 1:
                            di_last = decode2ed2w2e(&c2phbuf[0], &distg, (si<(MAX_GF_EL_1-iidx_mn1)?260:MAX_GF_EL_1), &erag, &scc_last, ob_mask, &bec_mask);
                        break;

                        case 2:
                            di_last = decode2ed2w3e(&c2phbuf[0], &distg, (si<(MAX_GF_EL_1-iidx_mn1)?260:MAX_GF_EL_1), &erag, &scc_last, ob_mask, &bec_mask);
                        break;

                        case 3:
                            di_last = decode2ed2w4e(&c2phbuf[0], &distg, (si<(MAX_GF_EL_1-iidx_mn1)?260:MAX_GF_EL_1), &erag, &scc_last, ob_mask, &bec_mask);
                            if (di != di_last)
                            {
                                if (erag==1)
                                    erasi1[0][eracnt1++] = MAX_INFO_INDEX -si; // hard erasure
                                else
                                    erasi1ss[0][eracnt1ss++]= MAX_INFO_INDEX-si; // soft erasure
                                if (di_last >= 0)
                                    cb[0][MAX_INFO_INDEX-si]=(unsigned int)(di_last);
                            }
                            else // count this as successful decode with 3 erasures
                            {
                                erasi1sss[0][eracnt1sss++]= MAX_INFO_INDEX-si;
                            }
                        } // end of switch scc_last
                    } // end of while

                    if (di == di_last)
                    {                           // start of addition of 21 Nov 2001
                        if (scc_last==2)
                        {
                            distg_last = distg;
                            di_last    = decode2ed2w3e(&c2phbuf[0], &distg, (si<(MAX_GF_EL_1-iidx_mn1)?260:MAX_GF_EL_1), &erag, &scc_last, ob_mask, &bec_mask);
                            if (di == di_last)
                            {
                                cb[0][MAX_INFO_INDEX-si]  = (unsigned int)(di);
                                scca               += scc_last -2;
                                sccx[scc_last-2]++;
                            }
                            else // di != di_last
                            {
                                di      = di_last;
                                di_last = decode2ed2w4e(&c2phbuf[0], &distg, (si<(MAX_GF_EL_1-iidx_mn1)?260:MAX_GF_EL_1), &erag, &scc_last, ob_mask, &bec_mask);
                                if (di != di_last)
                                {
                                    if (erag==1)
                                        erasi1[0][eracnt1++] = MAX_INFO_INDEX -si; // hard erasure
                                    else
                                        erasi1ss[0][eracnt1ss++] = MAX_INFO_INDEX-si; // soft erasure
                                    if (di_last >= 0)
                                        cb[0][MAX_INFO_INDEX-si] = (unsigned int)(di_last);
                                }
                                else // count this as successful decode with 3 erasures
                                {
                                    erasi1sss[0][eracnt1sss++]  = MAX_INFO_INDEX-si;
                                    cb[0][MAX_INFO_INDEX-si]          = (unsigned int)(di);
                                    scca                       += scc_last -1;
                                    sccx[scc_last-1]++;
                                }
                            }
                        } // end of scc_last == 2
                        else // end of addition of 21 Nov 2001, except for braces enclosing next three lines
                        {
                            cb[0][MAX_INFO_INDEX-si]  = (unsigned int)(di);
                            scca               += scc_last -1;
                            sccx[scc_last-1]++;
                        }
                    } // end of di == di_last  and scc_last != 2

                    //-------------------------------------------------------
                    if (uiDebugLevel > 50)
                    {
                        fprintf(stdout,"%3d_%1d_%1d d:%5.2f  ", di_last, scc_last-1, erag, distg);
                    }
                    //-------------------------------------------------------

                } // end of not too many out of bounds subcarrier values

            } // end of di<0 || scc>0

            else    // initial decode2 result >= 0 AND no subcarriers changed
            {
                cb[0][MAX_INFO_INDEX-si]  = (unsigned int)(di);
                scca                     += scc;
                sccx[scc]++;
            }

            //---------------------------------------------------------------
            if (uiDebugLevel > 50)
            {
                fprintf(stdout,"\n");
            }
            //---------------------------------------------------------------

        } // end of for si

        // merge raw blocks before decoding
        fbtow = 0;
        cbidx = MAX_INFO_INDEX;
        while (cbidx >= iidx_mn1)
        {
            cbmr[fbtow++] = cb[0][cbidx--];
        }

        hec    = eracnt1;
        sec    = eracnt1s;

        // save for diagnostic output
        ssec   = eracnt1ss;
        sssec  = eracnt1sss;
        ssssec = eracnt1ssss;

        //  use hard erasures first, then soft erasures, until maximum number
        //  of allowed erasures are used up, or all hard and soft erasures
        //  are used up
        //  Soft erasures should really be sorted first, so that "most likely
        //  to be wrong" symbols are erased first

        ulInputElRead = 0;
        while((eracnt1<k1) && (ulInputElRead<eracnt1s))
        {
            erasi1[0][eracnt1++] = erasi1s[0][ulInputElRead++];
        }

        ulInputElRead = 0;
        while((eracnt1<(k1- era_rsv[ec_mode])) && (ulInputElRead<eracnt1ss))
        {
            erasi1[0][eracnt1++] = erasi1ss[0][ulInputElRead++];
        }

        ulInputElRead = 0;
        while((eracnt1<(k1- era_rsv[ec_mode])) && (ulInputElRead<eracnt1sss))
        {
            erasi1[0][eracnt1++] = erasi1sss[0][ulInputElRead++];
        }

        //-------------------------------------------------------------------
        if (uiDebugLevel > 50)
        {
            for (si=0; si<=MAX_INFO_INDEX; si++)
                fprintf(stdout,"%3d ", cb[0][si]);
            fprintf(stdout,"\nerasure indexes: ");
            for (si=0; si<eracnt1; si++)
                fprintf(stdout,"%3d ", erasi1[0][si]);
            fprintf(stdout,"\n");
        }
        //-------------------------------------------------------------------

        // decode code block [0]
        eccnt1[0] = iDecode(&cb[0][0], eracnt1, &erasi1[0][0]);

        //-------------------------------------------------------------------
        if (uiDebugLevel > 50)
        {
            fprintf(stdout,"outer code correction count: %3d\n", eccnt1[0]);
            for (si=0; si<=MAX_INFO_INDEX; si++)
               fprintf(stdout,"%3d ", cb[0][si]);
            fprintf(stdout,"\n");
        }
        //-------------------------------------------------------------------

        // choose which merged blocks to use
        eci      = k1 - eracnt1;
        cbmerged = &cbmd[0];
        if (eci<0)
           cbmerged = &cbmr[0];
        else
        {
           if ((eci/2 + eracnt1) < eccnt1[0])
               cbmerged = &cbmr[0];
        }

        if (cbmerged == &cbmd[0])
        {
            // merge decoded blocks
            fbtow = 0;
            cbidx = MAX_INFO_INDEX;
            while (cbidx >= iidx_mn1)
            {
                cbmd[fbtow++] = cb[0][cbidx--];
            }
        }

        //-------------------------------------------------------------------
        if (uiDebugLevel > 0)
        {
            fprintf(stdout,"%3u    %4u   %3uh  %3us  %3ut  %3uf    %3d ",
               j, scca, hec, sec, ssec, sssec,  eccnt1[0]);
            if (cbmerged == &cbmd[0])
                ilc= '+';
            else
                ilc= '-';
            fprintf(stdout," || ");
            for (si=0; si<=NUM_REDUNDANT_SC; si++)
                fprintf(stdout," %3d", sccx[si]);
        }
        //-------------------------------------------------------------------

        if (j==0)    // 1st block is special case
        {
            //  extract header
            //  get file name, and file length
            //  open file, write data from first block

            cbidxm = 0;
            if (cbmerged[cbidxm++] != soh)
                {error_msg_to_stderr(ERN=19);}
            fnlen  = cbmerged[cbidxm++] ;
            if (fnlen > MAX_FILENAME_LENGTH)
                {error_msg_to_stderr(ERN=20); fnlen=MAX_FILENAME_LENGTH;}
            if (fnlen < 1)
                {error_msg_to_stderr(ERN=21);}
            file_lengthb  = cbmerged[cbidxm++];
            file_lengthb += cbmerged[cbidxm++]<<8;
            if (file_lengthb > MAX_FILE_LENGTH)
                {error_msg_to_stderr(ERN=22);}
            fbw = 0;

            // get output file name
            for (si=0; si<fnlen; si++)
            {
                tmpi = cbmerged[cbidxm++];
                if (tmpi== sod)
                    {cbidxm--; goto eofn;}
                if (tmpi > max_usr_sym)
                    {error_msg_to_stderr(ERN=23);}
                else
                    fnbuf[si] = (unsigned char)(tmpi);
            }
eofn:
            fnbuf[si] = 0x00;     // for string termination

            // for now, overwrite any existing file of the same name
            remove (fnbuf);
            FpOutputFile      = NULL;
            FpOutputFile      = fopen (&fnbuf[0], "wb");
            filename_OK_flag = 1;
            if (FpOutputFile==NULL)
            {
                filename_OK_flag = 0;
                FpOutputFile  = fopen (cDefaultOutputFilename, "wb");
                if (FpOutputFile==NULL)
                    {ERN=8; goto exit_l;}
            }
            if (cbmerged[cbidxm++] != sod)
                {error_msg_to_stderr(ERN=24);}
            fbtow = 0;
            while ((cbidxm < (MAX_GF_EL_1-iidx_mn1)) &&(fbtow<file_lengthb))
            {
                tmpi = cbmerged[cbidxm++];
                if (tmpi > max_usr_sym)
                    {error_msg_to_stderr(ERN=25); fotbuf[fbtow++]= 0x00; }
                else
                    fotbuf[fbtow++] = (unsigned char)(tmpi);
            } // end of while cbidx >= iidx_mn1

            fwr = fwrite(fotbuf, sizeof(unsigned char), fbtow, FpOutputFile);
            if (fwr!=fbtow)
                {error_msg_to_stderr(ERN=26);}
            if (ilc != '+')
                bad_blks++;
            fbw += fbtow;
            if (fbw >= file_lengthb)
                {fclose(FpOutputFile);

            //---------------------------------------------------------------
            if (uiDebugLevel > 0)
            {
                fprintf(stdout,"   %c\n", ilc);
            }
            //---------------------------------------------------------------

            goto normal_exit;
            }
        } // end of j==0
        else
        {
            // process remaining blocks
            cbidxm = 0;
            fbtow  = 0;
/*??*/      while ((cbidxm < (MAX_GF_EL_1 - iidx_mn1)) && ((fbw + fbtow) < file_lengthb))
            {
                tmpi = cbmerged[cbidxm++];
                if (tmpi > max_usr_sym)
                {
                    error_msg_to_stderr(ERN=25);
                    fotbuf[fbtow++] = 0x00;
                }
                else
                    fotbuf[fbtow++] = (unsigned char)(tmpi);
            }   // end of while cbidx >= iidx_mn1

            fwr = fwrite(fotbuf, sizeof(unsigned char), fbtow, FpOutputFile);
            if (fwr!=fbtow)
            {
                error_msg_to_stderr(ERN=26);
            }
            if (ilc != '+')
                bad_blks++;
            fbw += fbtow;
            if (fbw >= file_lengthb)
            {
                fclose(FpOutputFile);

                //-----------------------------------------------------------
                if (uiDebugLevel > 0)
                {
                    fprintf(stdout,"   %c\n", ilc);
                }
                //-----------------------------------------------------------

                if (j != (num_frsq-2))
                    ERN=27;
                goto normal_exit;
            }
        }

        //-------------------------------------------------------------------
        if (uiDebugLevel > 0)
        {
            fprintf(stdout,"   %c\n", ilc);
        }
        //-------------------------------------------------------------------

    } // end of for j (over each code block)

    //-----------------------------------------------------------------------
normal_exit:            // normal exit

    //-----------------------------------------------------------------------
    if (uiDebugLevel == 25)
    {
        fprintf (stdout, "\n\n// Cross Correlation Execution Count = %d\n", ulCCCount);
        fprintf (stdout, "// CC Execution Count 1 = %d\n",   ulCCCount1);
        fprintf (stdout, "// CC Execution Count 2 = %d\n",   ulCCCount2);
        fprintf (stdout, "// CC Execution Count 3 = %d\n",   ulCCCount3);
        fprintf (stdout, "// CC Execution Count 4 = %d\n",   ulCCCount4);
        fprintf (stdout, "// CC Execution Count 5 = %d\n",   ulCCCount5);
        fprintf (stdout, "// CC Execution Count 6 = %d\n\n", ulCCCount6);
    }
    //-----------------------------------------------------------------------

    if (filename_OK_flag == 1)
    {
        fprintf (stdout,"\nDecoded result is file: %s\n\n", &fnbuf[0]);
    }
    fprintf (stdout,"%3d bad blocks detected, out of %3d total blocks.\n", bad_blks, num_frsq-1);
    if (j == (num_frsq-1))
        ERN=0;

    //-----------------------------------------------------------------------
exit_l:
    for (j=0; j<NUM_TOTAL_SC; j++)
        free (fridxbetter[j]);
    for (j=0; j<NUM_TOTAL_SC; j++)
        free (fridxadj[j]);

    //-----------------------------------------------------------------------
exit_k:
    for (j=0; j<NUM_TOTAL_SC; j++)
        {free (dif7n[j]); free(dmamp[j]);}

    //-----------------------------------------------------------------------
exit_j1:
    remove(&scratch1[0]);
    remove(&scratch2[0]);
exit_j:
    fclose (FpInputFile);
    if (ERN != 0)
    {
        fprintf (stderr,"err_loc = %d ", err_loc);
        fprintf (stderr,"%s\n", erexmsg[ERN]);
    }
    if (ERN == 0)
    {
        if (bad_blks !=0)
             exit (bad_blks) ;
        else
            exit (0);
    }
    else
        exit (-ERN);

}
